from .StarList import StarList
from .file_helpers import *
import pandas as pd
from .fileformats import *


DAO_file_firstline = ' NL    NX    NY  LOWBAD HIGHBAD  THRESH     AP1  PH/ADU  RNOISE    FRAD'

def convert_dao_type(starlist, new_daotype, update_daotype=True):
        # type: (StarList, DAO.FType) -> bool
    """Converts from one daotype to another

    Only subset of conversions supported. You can also provide own map (directory) of column names"""
    for col in new_daotype.columns:
        if col not in starlist.columns:
            ct = _get_col_type(new_daotype.extension, col)
            if not ct.optional:
                if ct.default is None:
                    return False
                else:
                    starlist.loc[:, col] = ct.default # new col
    if update_daotype:
        starlist.DAO_type = new_daotype
    return True


def read_dao_file(file, dao_type = None):
    """
    Construct StarList from daophot output file.
    The header lines in file may be missing.
    :rtype: StarList
    :param file: open stream or filename, if stream dao_type must be specified
    :param dao_type: file format, one of DAO.XXX_FILE constants:
                    - DAO.COO_FILE
                    - DAO.AP_FILE
                    - DAO.LST_FILE
                    - DAO.NEI_FILE
                    - DAO.ALS_FILE
                If missing filename extension will be used to determine file type
                if file is provided as filename
    :return: StarList instance
    """
    ret = _parse_file(file, dao_type)
    return ret


def write_dao_file(starlist, file, dao_type=None, with_header=None):
    """
    Write StarList object into daophot  file.
    :param starlist: StarList instance to be writen
    :param file: writable stream or filename, if stream dao_type must be specified
    :param dao_type: file format, one of DAO.XXX_FILE constants:
                    - DAO.COO_FILE
                    - DAO.AP_FILE
                    - DAO.LST_FILE
                    - DAO.NEI_FILE
                    - DAO.ALS_FILE
                If missing extension of file will be used to determine file type
                if file is provided as filename
    :param with_header: True, False or None. If None header will be written if not None in starlist
    :rtype: None
    """
    if dao_type is None:
        if starlist.DAO_type is None:
            raise Exception('Can not determine file format')
        dao_type = starlist.DAO_type
    elif dao_type != starlist.DAO_type:
        converted = convert_dao_type(starlist, dao_type, update_daotype=False)
        if not converted:
            raise Exception('Can not convert columns {} into {} '.format(starlist.columns, dao_type.columns))
    _write_file(starlist, file, dao_type, with_header=with_header)

def _get_col_type(file_ext, column):
    # type: (str, str) -> DAO.CType
    coltype = DAO.columns.get((file_ext, column))  # lookup for (filetype,col)
    if not coltype:
        coltype = DAO.columns.get(column)  # lookup for col
    if not coltype:
        coltype = DAO.columns['_default']
    return coltype


def _write_file(starlist, file, dao_type, with_header=None):
    f, to_close = get_stream(file, 'w')
    if with_header != False:
        if starlist.DAO_hdr is not None:
            starlist.DAO_hdr['NL'] = dao_type.NL
            write_dao_header(starlist.DAO_hdr, f)
            f.write('\n')
        elif with_header:
            raise Exception('Attempt to save dao file with header from StarList with DAO_hdr=None')
    _write_table(starlist, f, dao_type)
    close_files(to_close)


def dump_dao_hdr(hdr, line_prefix=''):
    """
    returns two line string representation of header dictionary
    :param dict hdr: dao header dictionary like StarList.DAO_hdr
    :param str line_prefix: add this prefix at beginning of every line (e.g. comment char)
    :rtype:str
    """
    first_line = ''
    second_line = ''
    second_line_frmt = ['{:3.0f}', '{:6.0f}', '{:6.0f}', '{:8.1f}', '{:8.1f}', '{:8.2f}',
                        '{:8.2f}', '{:8.2f}', '{:8.2f}', '{:8.2f}']
    second_line_frmt.reverse()
    try:
        for token in DAO_file_firstline.split(' '):
            if token != '':
                val = hdr[token]
                first_line += token
                fmt = second_line_frmt.pop()
                second_line += fmt.format(float(val))
            first_line += ' '
    except KeyError:
        # sometimes header is shorter, it's OK'
        pass
    return line_prefix + first_line + '\n' + line_prefix + second_line + '\n'


def write_dao_header(hdr, stream, line_prefix=''):
    """
    writes two lines of dao header
    :param dict hdr: dao header dictionary like StarList.DAO_hdr
    :param file stream: to write
    :param str line_prefix: add this prefix at beginning of every line (e.g. comment char)
    """
    f, to_close = get_stream(stream, 'w')
    stream.write(dump_dao_hdr(hdr, line_prefix=line_prefix))
    close_files(to_close)


def _write_table(starlist, file, dao_type):
    # preapre columns (from daotype in order but only existing in starlist)
    pd.options.mode.chained_assignment = None  # default='warn'
    columns = [c for c in dao_type.columns if c in starlist.columns]
    coltypes = [_get_col_type(dao_type.extension, c) for c in columns]
    towrite = starlist[columns]

    for i, row in towrite.iterrows():
        for col, coltype, val in zip(columns, coltypes, row):
            if pd.isnull(val):
                val = coltype.NaN[0]
            file.write(coltype.format.format(val))
        file.write('\n')

def _parse_file(file, dao_type):
    if dao_type is None and isinstance(file, str):
        _, ext = os.path.splitext(file)
        dao_type = DAO.file_types.get(ext)

    f, to_close = get_stream(file, 'r')
    try:
        hdr, _ = read_dao_header(f)
        fl = _parse_table(f, hdr, dao_type)
    finally:
        close_files(to_close)
    fl.DAO_hdr = hdr
    return  fl


def read_dao_header(stream, line_prefix=''):
    """
    tries to read dao header, if fails returns already read characters
    :param file stream: open input file
    :param line_prefix: additional prefix expected on the beginning of line
    :return: tuple (header dict, stolen chars)
             if header is detected, reads 2 lines of stream and returns (dict, None)
             else reads couple of chars and return (None, couple-of-chars)
    """
    # we are very smart, if first two characters of ile are not ' N', we suppose there is no header
    # and we get not too much to disturb further parsing of tables
    signature = ' NL'
    stolen_chars = ''
    for c in line_prefix + signature:
        r = stream.read(1)
        stolen_chars += r
        if c != r: # no header signature  found
            return None, stolen_chars

    hdr = stolen_chars + stream.readline() # first line
    val = stream.readline()
    return parse_dao_hdr(hdr, val, line_prefix), None


def parse_dao_hdr(hdr, val, line_prefix=''):
    """
    creates dao header dict form two lines of file header
    :param str hdr: first line
    :param str val: second line
    :param line_prefix: expected line prefix
    :return: dict with dao header compatible with StarList.DAO_header
    """
    # cut prefix
    hdr = hdr[len(line_prefix):] # first line
    val = val[len(line_prefix):] # second line
    # split keys, split values, zip them into list of tuples, then create dictionary out of that
    return dict(zip(hdr.split(), val.split()))


def _parse_table(f, hdr, dao_type):
    if dao_type is not None and dao_type.read_cols is not None:  # limit number of read cols
        df = pd.read_table(f, header=None, sep='\s+', usecols=range(dao_type.read_cols))
    else:
        df = pd.read_table(f, header=None, sep='\s+')

    #df.insert(0, 'id', df.index.to_series())
    if dao_type is None:
        dao_type = _guess_filetype(hdr, df)
    if dao_type == DAO.AP_FILE:  # two row per star format correction
        odd = df.iloc[0::2]
        odd.columns = DAO.AP_FILE_ODD.columns[:odd.columns.size]
        even = df.iloc[1::2]
        even.columns = DAO.AP_FILE_EVEN.columns[:even.columns.size]
        even.index = odd.index
        df = odd.join(even, rsuffix='foo')
    else:
        df.columns = dao_type.columns[:df.columns.size]


    df.id = df.id.astype(int)
    df.index = df.id

    # find NaN
    for col in df.columns:
        coltype = _get_col_type(dao_type.extension, col)
        if coltype.NaN:
            df[col].replace(coltype.NaN, pd.np.nan, inplace=True)

    ret = StarList(df)
    ret.DAO_type = dao_type
    return ret


def _guess_filetype(header, table):
    type = DAO.UNKNOWN_FILE
    if header:
        colno = table.columns.size
        NL = int(header['NL'])
        type = DAO.UNKNOWN_FILE
        if NL == 1:
            if colno == 7:
                type = DAO.COO_FILE
            elif colno == 9:
                type = DAO.ALS_FILE
            elif colno == 5:
                type = DAO.SHORT_FILE
        elif NL == 2:
            type = DAO.AP_FILE
        elif NL == 3:
            if colno == 6:
                type = DAO.LST_FILE
            elif colno == 5:
                type = DAO.NEI_FILE
    return type



# coding=utf-8
"""Greenstick's code form http://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
encapsulated into class
"""

import sys


class ProgressBar(object):
    def __init__(self, total=100, prefix='', suffix='', decimals=1, bar_length=20, step=1, cleanup=True):
        """
        Create terminal progress bar 
        :param int total: total iterations (Int)
        :param str prefix: prefix string (Str)
        :param str suffix: suffix string (Str)
        :param int decimals: positive number of decimals in percent complete (Int)
        :param int bar_length: character length of bar (Int)
        :param int step: allows automatic progress increasing on parameter-less print_progress call
        :param bool cleanup: if true (default) erase progressbar when completed 
        """
        self.total = total
        self.prefix = prefix
        self.suffix = suffix
        self.decimals = decimals
        self.bar_length = bar_length
        self.step = step
        self.iteration = 0
        self.cleanup = cleanup

    def print_progress(self, iteration=None):
        """
        Call in a loop to print terminal progress bar
        @params:
            iteration 
        """
        if iteration is None:
            iteration = self.iteration + self.step
        if iteration > self.total:
            iteration = self.total
        format_str = "{0:." + str(self.decimals) + "f}"
        percents = format_str.format(100 * (iteration / float(self.total)))
        filled_length = int(round(self.bar_length * iteration / float(self.total)))
        bar = '■' * filled_length + '·' * (self.bar_length - filled_length)
        stream = sys.stderr
        stream.flush()
        stream.write('\r%s [%s] %s%s %s' % (self.prefix, bar, percents, '%', self.suffix)),
        if iteration == self.total:
            if self.cleanup:
                self.clean_progress()
            else:
                stream.write('\n')
            stream.flush()
        self.iteration = iteration

    def clean_progress(self):
        n = len(self.prefix) + len(self.suffix) + self.bar_length + self.decimals + 10
        stream = sys.stderr
        stream.flush()
        stream.write('\r' + ' '*n + '\r'),



from __future__ import absolute_import, division, print_function

__metaclass__ = type
# see https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef
from os.path import join, basename
from astwro.exttools import Runner
from astwro.config import find_opt_file

#TODO: suport for configuration and conf files (xommon with scamp)


class SexResults(object):

    def __init__(self, dir, results_filename):
        super(SexResults, self).__init__()
        self.dir = dir
        self.stars_file = results_filename
        self._stars = None

    @property
    def stars(self):
        if self._stars is None:
            import astropy.io.fits as pyfits
            self._stars = pyfits.getdata(join(self.dir, self.stars_file),2)
        return self._stars


class Sextractor(Runner):
    """ Emmanuel BERTIN `sextractor` runner

    Object of this class maintains single process of `extractor` and it's working directory.

    Note
    ----
    Very `alpha` stage, very limited


    Parameters
    ----------
    dir : str, TmpDir, optional
        Working directory for sextractor process
        Default - new TmpDir will be created and destroyed with Sextractor object
    conf : str, optional
        Patch to sextractor's configuration file, Default: package default file
    param : str, optional
        Patch to sextractor's parameters file (specified in conf), Default: package default file
    conv : str, optional
        Patch to sextractor's convolution kernel file (specified in conf), Default: package default file
    """

    def __init__(self, dir=None, conf=None, param=None, conv=None, output=None):
        # base implementation of __init__ calls `_reset` also
        super(Sextractor, self).__init__(dir=dir, batch=False)
        self.conf_file = conf
        self.param_file = param
        self.conv_file = conv
        self.output_file = output
        self.SEXresults = None
        self._update_executable('sextractor')

    def _reset(self):
        super(Sextractor, self)._reset()

    def __deepcopy__(self, memo):
        from copy import deepcopy
        new = super(Sextractor, self).__deepcopy__(memo)

        return new

    def _pre_run(self, wait):
        super(Sextractor, self)._pre_run(wait=wait)



    def _init_workdir_files(self, dir):
        super(Sextractor, self)._init_workdir_files(dir)

    def __call__(self, image):
        """ Run sextractor on image

        Parameters
        ----------
        image : str
            Path of the FITS image

        Returns
        -------
        SexResults object
        """
        if self.conf_file is None:
            conf_file = find_opt_file('sextractor.conf', package='sex')
            out_file = 'output1.fits'
            self._prepare_input_file(find_opt_file('sextractor.param', package='sex'), preservefilename=True)
            self._prepare_input_file(find_opt_file('sextractor.conv', package='sex'), preservefilename=True)
        else:
            out_file  = self.output_file
            conf_file = self.conf_file
            if self.param_file is not None:
                self._prepare_input_file(self.param_file, preservefilename=True)
            if self.conv_file is not None:
                self._prepare_input_file(self.conv_file, preservefilename=True)
        self._prepare_input_file(conf_file, preservefilename = True)
        self._prepare_input_file(image, preservefilename=True)

        self.SEXresults = SexResults(dir=self.dir.path,  results_filename=out_file)
        self.arguments = [basename(image), '-c', basename(conf_file)]
        self.run()
        return self.SEXresults






from .StarList import StarList
from .file_helpers import *
from .daofiles import parse_dao_hdr, write_dao_header, DAO_file_firstline, DAO
from .file_helpers import as_starlist
import pandas as pd
import re

_ds9_regexp = re.compile(
    r'[+-]? *circle[( ] *([+-]?\d+[.]?\d*) *[, ] *([+-]?\d+[.]?\d*).+#.*id *= *(\d+)')
_ds9_wcs_regexp = re.compile(
    r'[+-]? *circle[( ] *([+-]?\d+:\d+:\d+[.]?\d*) *[, ] *([+-]?\d+:\d+:\d+[.]?\d*).+#.*id *= *(\d+)')
_ds9_no_id_regexp = re.compile(
    r'[+-]? *circle[( ] *([+-]?\d+[.]?\d*) *[, ] *([+-]?\d+[.]?\d*)')
_ds9_no_id_wcs_regexp = re.compile(
    r'[+-]? *circle[( ] *([+-]?\d+:\d+:\d+[.]?\d*) *[, ] *([+-]?\d+:\d+:\d+[.]?\d*)')
_ds9_system_wcs = re.compile('fk4|fk5|J2000|B1950|ICRS', re.IGNORECASE)
_ds9_system_xy = re.compile('PHYSICAL|IMAGE', re.IGNORECASE)


def read_ds9_regions(file):
    # type: (object) -> StarList
    """
    Reads ds9 region
    :param file: filename or open input stream
    :return: StarList object

    Returned object has columns id, x, y, auto_id

    Boolean column auto_id indicates weather id for item is read from file (#id=xxx comment) or
    generated by function.
    """

    f, to_close = get_stream(file, 'rt')
    # s = StarList.new()
    data = []
    data_noid = []
    dao_hdr1 = None
    hdr = None
    sys_wcs, sys_xy = (1,2)
    system = None
    for line in f:
        if line[0] == '#':
            if line[1:11] == DAO_file_firstline[:10]:  # dao header found in comment
                dao_hdr1 = line
                continue
            if dao_hdr1 is not None:  # second line of dao header
                hdr = parse_dao_hdr(dao_hdr1, line, '#')
        else:
            if system is None:
                if _ds9_system_wcs.search(line):
                    system = sys_wcs
                elif _ds9_system_xy.search(line):
                    system = sys_xy
                pass
            m = _ds9_regexp.search(line)
            if m is not None:  # s[id] = (id, x, y)
                data.append([int(m.group(3)), float(m.group(1)), float(m.group(2))])
            else:
                m = _ds9_wcs_regexp.search(line)
                if m is not None:  # s[id] = (id, ra, dec)
                    data.append([int(m.group(3)), str(m.group(1)), str(m.group(2))])
                else:
                    m = _ds9_no_id_regexp.search(line)  # s[?] = (x, y)
                    if m is not None:
                        data_noid.append([float(m.group(1)), float(m.group(2))])
                    else:
                        m = _ds9_no_id_wcs_regexp.search(line)  # s[?] = (ra, dec)
                        if m is not None:
                            data_noid.append([str(m.group(1)), str(m.group(2))])

        dao_hdr1 = None
    close_files(to_close)
    if system == sys_wcs:
        s = StarList(data, columns = ['id', 'ra', 'dec'])
        s_noid = StarList(data_noid, columns=['ra', 'dec'])
    else:
        s = StarList(data, columns = ['id', 'x', 'y'])
        s_noid = StarList(data_noid, columns = ['x', 'y'])
    s.index = s['id']
    s['auto_id'] = False
    if not s_noid.empty:
        id_starts_from = 1 if s.empty else s.id.max() + 1
        ids = range(id_starts_from, id_starts_from + s_noid.stars_number())
        s_noid['id'] = ids
        s_noid.index = ids
        s_noid['auto_id'] = True
        if s.empty:
            s = s_noid
        else:
            s = s.append(s_noid)

    s.DAO_hdr = hdr
    s.DAO_type = DAO.RADEC_FILE if system == sys_wcs else DAO.XY_FILE
    return s


def write_ds9_regions(starlist, filename,
                      color='green', width=1, size=None, font=None, label='{id:.0f}',
                      exclude=None, indexes=None, colors=None, sizes=None, labels=None,
                      color_column=None, size_column=None,
                      comment=None, add_global=None, WCS=False):
    """
    Writes ds9 region file.
    Some regions can be visually distinguish by providing additional indexes to select those regions
    with specific attributes
    :param StarList starlist: StarList object to dump
    :param str filename:      output filename or stream open for writing
    :param str color:         default color
    :param int width:         default line width
    :param int size:          default radius (default 8px or 2")
    :param str font:          ds9 font specification e.g. "times 12 bold italic"
    :param str label:         format expression for label, use col names
    :param pd.Index exclude:  index of disabled regions, if None all are enabled
    :param [pd.Index] indexes: additional indexes to include specific color and size attributes
    :param [str] colors:      specific colors for indexes
    :param [int] sizes:       specific sizes for indexes
    :param [str] labels:      specific labels for indexes
    :param str color_column:  column of starlist with color values
    :param str size_column:   column of starlist with size values
    :param str add_global:    content of additional 'global' if not None
    :param str comment:       content of additional comment line if not None
    :param bool or str WCS:   If true, columns `ra` and `dec` will be used and coord system set to ICRS
                              If nonepmpty string, string will be used as system description
                              If None, False or '', columns 'x','y' will be used and system set to IMAGE
    Example:
    write_ds9_regions(sl, 'i.reg', color='blue',
                        indexes=[saturated, psf],
                        colours=['yellow', 'red'],
                        sizes=[12, None],
                        labels=[None, 'PDF:{id}'],
                        exclude=faint)
    Generates regions file i.reg of blue circles, radius 8,
    objects present in index saturated will have larger yellow circles
    objects present in index psf will be red and labeled with prefix PSF:
    objects present in index faint will be disabled by '-' sign and not displayed by ds9, but can be parsed back
    """
    if WCS:
        xcol = 'ra'
        ycol = 'dec'
        starlist = as_starlist(starlist)
    else:
        xcol = 'x'
        ycol = 'y'
        starlist = as_starlist(starlist, updateskycoord=False)
    try: (starlist[xcol], starlist[ycol])
    except KeyError as e:
        raise KeyError('No coordinate columns ({},{}) in starlist. Check WCS parameter also'.format(xcol, ycol))
    f, to_close = get_stream(filename, 'w')
    f.write('# Region file format: DS9 version 4.0\n')
    if starlist.DAO_hdr is not None:
        write_dao_header(starlist.DAO_hdr, f, '#')
    if comment is not None:
        f.write('#{}\n'.format(comment))
    if color is not None:
        f.write('global color={}\n'.format(color))
    if width is not None:
        f.write('global width={}\n'.format(width))
    if font is not None:
        f.write('global font={}\n'.format(font))
    if add_global is not None:
        f.write('global {}\n'.format(add_global))
    if not WCS:
        f.write('image\n')
    else:
        system = WCS if isinstance(WCS, str) else 'icrs'
        f.write(system+'\n')
    for i, row in starlist.iterrows():
        if exclude is not None and i in exclude:
            f.write('-')
        if size is not None:
            s = size
        else:
            s = '2"' if WCS else 8
        text = label.format(**row)
        c = ''
        if size_column is not None:
            s = row[size_column]
        if color_column is not None:
            c = ' color=' + row[color_column]
        if indexes is not None:
            for n in range(len(indexes)):
                if i in indexes[n]:
                    if sizes and sizes[n] is not None:
                        s = sizes[n]
                    if colors and colors[n] is not None:
                        c = ' color=' + colors[n]
                    if labels and labels[n] is not None:
                        text = labels[n].format(**row)
        f.write('circle({},{},{}) #{} text="{}" id={:d}\n'.format(row[xcol], row[ycol], s, c, text, i))
    close_files(to_close)



from .TmpDir import TmpDir
from .CycleFile import CycleFile
from .ProgressBar import ProgressBar
from .fits_list import make_fits_table

def tmpdir(use_existing=None, prefix='astwro_tmp_', base_dir=None):
    """
    Creates instance of TmpDir which creates and keeps lifetime of temporary directory
    :param str use_existing:    If provided, instance will point to that directory and not delete it on destruct
    :param str prefix:          Prefix for temporary dir
    :param str base_dir:        Where to crate tem dir, in None system default is used
    :rtype: TmpDir
    """
    return TmpDir(use_existing, prefix, base_dir)


def cyclefile(path, basename, extension='', create_symlinks=True, symlink_suffix='_last', auto_close=True):
    """
    Creates CycleFile which can create series of files with names containing counter,
    with symlink to newest one
    :param str path: patch to the file, absolute or relative, can be empty string '' for current directory
    :param str basename: first part of file name before counter
    :param str extension:  part of filename after counter
    :param str create_symlinks: whether to create symlink to newest file
    :param str symlink_suffix:  part of suffix filename in place of counter value
    :param bool auto_close: close prev files on next_file nad destruction
    :rtype: CycleFile
    """
    return CycleFile(path=path, basename=basename, extension=extension, create_symlinks=create_symlinks,
                     symlink_suffix=symlink_suffix, auto_close=auto_close)


def progressbar(total=100, prefix='', suffix='', decimals=1, bar_length=20, step=0):
    """
    Creates Text console progress bar object, print progress using print_progress() method
    :param int total: total iterations (Int)
    :param str prefix: prefix string (Str)
    :param str suffix: suffix string (Str)
    :param int decimals: positive number of decimals in percent complete (Int)
    :param int bar_length: character length of bar (Int)
    :param int step: allows automatic progress increasing on parameter-less print_progress call
    :rtype:ProgressBar
    """
    return ProgressBar(total=total, prefix=prefix, suffix=suffix, decimals=decimals, bar_length=bar_length, step=step)



from __future__ import print_function
from astwro.pydaophot import daophot, allstar, fname
from astwro.starlist import read_dao_file
from astwro.utils import tmpdir
import os
import sys
import numpy as np

basefilename = 'i'
results_path = None
pick_candidates = 80
pick_mag_limit = 20
run_parallel = 8

#fits = basefilename + '.fits'
fits = 'NGC6871.fits'
coo = basefilename + '.coo'
ap = basefilename + '.ap'
lst = basefilename + '.lst'
als = basefilename + '.als'

dopt = 'daophot.opt'
popt = 'photo.opt'
aopt = 'allstar.opt'

if not os.path.isfile(dopt):
    dopt = None
if not os.path.isfile(popt):
    popt = None
if not os.path.isfile(aopt):
    aopt = None

results_dir = tmpdir(use_exiting=results_path, prefix='psf_red_', base_dir=os.getcwd())
results_dir.dir_is_tmp = False
print ('Expect results in: {!s}'.format(results_dir),file=sys.stderr)

dp = daophot(image_file=fits, daophotopt=dopt, photoopt=popt, dir=results_dir)

# 1. Make sure that we are ready to PSF photometry, collect all needed files in dp.dir
# --

# FI needed?
if not os.path.isfile(coo):
    dp.FInd(1,1)
    assert (not os.path.isfile(ap) and not os.path.isfile(lst))
else:
    dp.copy_to_runner_dir(coo, fname.COO_FILE)

# PH needed?
if not os.path.isfile(ap):
    dp.PHotometry()
    assert (not os.path.isfile(lst))
else:
    dp.copy_to_runner_dir(ap, fname.AP_FILE)

# PI needed?
if not os.path.isfile(lst):
    dp.PIck(pick_candidates, faintest_mag=pick_mag_limit)
else:
    dp.copy_to_runner_dir(lst, fname.LST_FILE)

# run tasks list
dp.run(wait=True)

# 2. Now th game begins. our population has one runner, no PSF performed so we not call it the best
# ---
# runner runs only once, create new with old one's directory
population = [daophot(dir=dp.dir)]
prev_winner = winner = None
prev_winner_score = 1001
winner_score = 1000
log = []
counter = 0

# score function - best score wins in population
score_function = lambda (result): result['als_chi']

while winner_score < prev_winner_score:
    prev_winner = winner
    prev_winner_score = winner_score
    # population ready for psf photometry
    # because image file was not provided in constructor, AT must be called explicitly
    for d in population:
        d.ATtach()
        d.PSf()
    # run them in parallel, but not at once
    for n in range(0, len(population), run_parallel):
        for d in population[n:n+run_parallel]:  # run group
            d.run(wait=False)
        for d in population[n:n+run_parallel]:  # wait for group
            d.wait_for_results()
            # print (d.output)
    # time for allstar
    population_alls = [allstar(dir=d.dir, allstaropt=aopt, create_subtracted_image=False) for d in population]
    # run them same way
    for n in range(0, len(population_alls), run_parallel):
        for d in population_alls[n:n + run_parallel]:  # run group
            d.run(wait=False)
        for d in population_alls[n:n + run_parallel]:  # wait for group
            d.wait_for_results()
            print (d.output)
    # measure results and chose the best
    scores = np.zeros_like(population)
    for i, d in enumerate(population):
        conv_stars = read_dao_file(d.file_from_runner_dir(fname.ALS_FILE))
        psf_stars = read_dao_file(d.file_from_runner_dir(fname.LST_FILE))
        result = {
            'n': len(log),
            'conv_stars': len(conv_stars), # converged by allstar
            'psf_stars': len(psf_stars),   # pdf stars
            'als_chi': conv_stars['chi'].mean(),  # mean of chi form ALS
            'psf_chi': d.PSf_result.chi,    # chi from daophot PS
            'removed_star': 0
        }
        scores[i] = score_function(result)
        log.append(result)
        if counter % 8 == 0:
            print ('Done:{:5d}'.format(counter))
        counter += 1
    # and the winner is....
    max_scored = scores.argmax()
    winner_score = scores[max_scored]
    winner = population[max_scored]
    print ('Round winner scored {:f} after {:d} calculations')

    winner.copy_from_runner_dir(fname.LST_FILE, os.path.join(results_dir, 'round'))
    winning_psf_stars = read_dao_file(winner.file_from_runner_dir(fname.LST_FILE))

    # now prepare next population
#    population = [daophot(dir=winner.dir.clone()) for _ in ]

    ## calculate score for population:
    # define score function

    ## TODO metadane sesji - individual, population, round

    break ## TODO: temporary break




def fits_image():
    return __sampledata('NGC6871.fits')


def coo_file():
    return __sampledata('NGC6871.coo')


def lst_file():
    return __sampledata('NGC6871.lst')


def ap_file():
    return __sampledata('NGC6871.ap')


def psf_file():
    return __sampledata('NGC6871.psf')


def als_file():
    return __sampledata('NGC6871.als')


def __sampledata(filename):
    import os
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)



from .Daophot import Daophot
from .Allstar import Allstar
#from .ASRunner import ASRunner
#from .dao import allstar, daophot, daophot_cfg
from ._version import __version__, __version_info__

#__all__ = ['allstar', 'daophot', 'daophot_cfg']




#! /usr/bin/env python
# coding=utf-8
""" Grep-like tool for FITS headers

    Call commandline: ``grepfitshdr --help`` for parameters info.
"""
from __future__ import print_function, division

from sys import stdout, stdin

import astwro.starlist as sl
import astwro.tools.__commons as commons



def __do(arg):
    s = sl.read_dao_file(stdin)
    sl.write_ds9_regions(s, stdout, color=arg.color, width=arg.width, size=arg.size)


def __arg_parser():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='creates ds9 region out of stars list\n'
                    'star list on stdin, region file on stdout')
    parser.add_argument('-c', '--color', type=str, default='green',
                        help='color in format accepted by ds9, default: green')
    parser.add_argument('-r', '--size', type=float, default=8.0,
                        help='radius of object circle, default: 8')
    parser.add_argument('-w', '--width', type=int, default=1,
                        help='line width, default 1')
    parser.add_argument('-i', '--id-col', type=int, default=1,
                        help='number of column in star list containing object id, default: 1')
    parser.add_argument('-x', '--x-col', type=int, default=2,
                        help='number of column in star list containing x coordinate, default: 2')
    parser.add_argument('-y', '--y-col', type=int, default=3,
                        help='number of column in star list containing y coordinate, default: 3')
    return parser


def main(pattern, file, **kwargs):
    """Entry point for python script calls. Parameters identical to command line"""
    # Extract default arguments from command line parser and apply kwargs parameters
    args = commons.bunch_kwargs(__arg_parser(), positional=[pattern, file], **kwargs)
    # call main routine - common form command line and python calls
    return __do(args)


def info():
    """Prints commandline help message"""
    commons.info(__arg_parser())


def commandline_entry():
    # Entry point for command line
    __args = __arg_parser().parse_args()  # parse command line arguments
    n = __do(__args)  # call main routine - common form command line and python calls
    return  0

if __name__ == '__main__':
    code = commandline_entry()
    exit(code)



# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import unittest
from astwro.sex import Sextractor
from astwro.sex import Scamp
from astwro.sampledata import fits_image

# TODO: switch to pytest

class TestRunners(unittest.TestCase):

    def test_sextractor_run(self):
        sex = Sextractor()
        sex.run()

        # self.assertAlmostEqual(f.sky, s.sky)
        # self.assertAlmostEqual(f.skydev, s.skydev)
        # self.assertAlmostEqual(f.mean, s.mean)
        # self.assertAlmostEqual(f.median, s.median)
        # self.assertAlmostEqual(f.pixels, s.pixels)


    def test_astometry(self):
        sex = Sextractor()
        scamp = Scamp()
        sexresult = sex(fits_image())
        self.assertGreater(len(sexresult.stars), 0)
        scampresult = scamp(sexresult)





suite = unittest.TestLoader().loadTestsFromTestCase(TestRunners)
if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)


# coding=utf-8
from __future__ import absolute_import, division, print_function, generators

__metaclass__ = type

import numpy as np
import re
import matplotlib.pyplot as plt
from astwro.phot import DiffPhot


class Plottable(object):
    plot_styles =    ['alpha', 'ls', 'linestyle', 'c', 'color', 'lw', 'linewidth',
                      'marker', 'mec', 'mew', 'mfc', 'ms', 'zorder', 'label']
    scatter_styles = ['alpha', 'c', 's', 'marker', 'edgecolors', 'label']
    (tool_plot, tool_errorbar, tool_scatter) = (1,2,3)
    def __init__(self, **kwargs):
        super(Plottable, self).__init__()
        _kwargs = kwargs
        # plot_styles =    {s: None for s in 'alpha', 'ls', 'c', 'lw', 'marker', 'mec', 'mew', 'mfc', 'ms', 'zorder'}
        # scatter_styles = {s: None for s in 'alpha', 'c', 's', 'marker', 'edgecolors'}

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = self._setup_axes(**kwargs)
        self._plotonaxies(ax, **kwargs)
        return ax

    @staticmethod
    def _setup_axes(**kwargs):
        fig = plt.figure()
        return fig.add_subplot(111)

    def _plotonaxies(self, ax, **kwargs):
        pass

    @staticmethod
    def _style_kwargs(style, kwargs_list):
        res = None
        for kwargs in kwargs_list:
            try: res = kwargs[style]
            except KeyError: pass
        return res

    @staticmethod
    def _styles_kwargs(styles_list, kwargs_list):
        res = {}
        for s in styles_list:
            kw = Plottable._style_kwargs(s, kwargs_list)
            if kw is not None:
                res[s] = kw
        return res

    @staticmethod
    def _plot_styles_kwargs(kwargs_list):
        return Plottable._styles_kwargs(Plottable.plot_styles, kwargs_list)

    @staticmethod
    def _scatter_styles_kwargs(kwargs_list):
        return Plottable._styles_kwargs(Plottable.scatter_styles, kwargs_list)


class FigurePlotter(Plottable):
    def __init__(self, **kwargs):
        super(FigurePlotter, self).__init__(**kwargs)

    def _plotonaxies(self, ax, **kwargs):
        super(FigurePlotter, self)._plotonaxies(ax, **kwargs)
        r = kwargs.get('xlim')
        if r is not None:
            ax.set_xlim(r[0], r[1])
        r = kwargs.get('ylim')
        if r is not None:
            ax.set_ylim(r[0], r[1])


class Layer(Plottable):
    def __init__(self, **kwargs):
        super(Layer, self).__init__(**kwargs)
        self.default_styles = {}

    def plot_styles_kwargs(self, **kwargs):
        return self._plot_styles_kwargs([self.default_styles, kwargs])

    def scatter_styles_kwargs(self, **kwargs):
        return self._scatter_styles_kwargs([self.default_styles, kwargs])

    def _plotonaxies(self, ax, **kwargs):
        pass

class ShiftedLayer(Layer):
    def __init__(self, dx, dy, **kwargs):
        super(ShiftedLayer, self).__init__(**kwargs)
        self.dx = dx
        self.dy = dy


class IsochromeLayer(ShiftedLayer):
    def __init__(self, bands, age, Av, distance, feh=0.001, maxm=18.25, **kwargs):
        super(IsochromeLayer, self).__init__(**kwargs)
        from isochrones.mist import MIST_Isochrone
        MIST = MIST_Isochrone(bands)
        iso  = MIST.isochrone(age=np.log10(age), feh=feh, AV=Av, distance=distance, maxm=maxm)

    def _plotonaxies(self, ax, **kwargs):
        super(IsochromeLayer, self)._plotonaxies(ax, **kwargs)




class CollectionLayer(Layer):
    operators = {
        '-': lambda x, y: x - y,
        '+': lambda x, y: x + y,
    }
    oper_reg = re.compile('\s*(\w+)\s*([+-])\s*(\w+)\s*')

    def __init__(self, collections, masks=None, tool=Plottable.tool_plot, colors=None, sizes=None, subsets=None, **kwargs):
        super(CollectionLayer, self).__init__(**kwargs)
        self.collections = collections
        self.masks = masks
        self.default_styles['ls'] = 'None'
        self.default_styles['marker'] = '.'
        self.default_styles.update(self._plot_styles_kwargs([kwargs]))
        self.tool = tool
        self.colors = colors
        self.sizes = sizes
        self.subsets = subsets

    def __len__(self):
        return len(list(self.collections.values())[0])

    def get_values(self, color):
        try:
            m = self.oper_reg.match(color)
            if not m:
                return self.collections[color]
            else:
                a = m.groups()[0]
                op = m.groups()[1]
                b = m.groups()[2]
                return self.operators[op](self.collections[a], self.collections[b])
        except (TypeError, IndexError, KeyError):
            raise ValueError('"color" should be a band name or two color +/- operation, eg. "V", "B-V"')

    def get_masks(self, **kwargs):
        masks = kwargs.get('masks')
        n = len(self)
        if masks is None:
            masks = [np.ones(n, dtype=bool)]
        for i, mask in enumerate(masks):
            if (len(mask) != n):
                boolmask = np.zeros(n, dtype=bool)
                boolmask[mask] = True
                masks[i] = boolmask
        return masks

    def _plotonaxies(self, ax, **kwargs):
        super(CollectionLayer, self)._plotonaxies(ax, **kwargs)
        x = self.get_values(kwargs['xcolor'])
        y = self.get_values(kwargs['ycolor'])
        masks = self.get_masks(**kwargs)
        if self.tool == self.tool_plot:
            styles = self.plot_styles_kwargs(**kwargs)
            for mask in masks:
                ax.plot(x[mask], y[mask], **styles)
        elif self.tool == self.tool_scatter:
            styles = self.scatter_styles_kwargs(**kwargs)
            if self.colors:
                if self.colors == True: # just bool, cycle std cycler over subsets
                    assert len(self.subsets) == len(x), 'If `colors == True` for scatter, `subsets` of size of data should be specified'
                    palette = self.get_scatter_cycle_palette()
                    N = len(palette)
                    c = [palette[i%N] for i in self.subsets]
                else:
                    c = self.colors
                styles['c'] = c
            if self.sizes:
                styles['s'] = self.sizes
            for mask in masks:
                ax.scatter(x[mask], y[mask], **styles)

    def get_scatter_cycle_palette(self): return ['C'+l for l in ['0123456789']]


class CompoundPlot(FigurePlotter):
    def __init__(self, layers, **kwargs):
        super(CompoundPlot, self).__init__(**kwargs)
        self.layers = layers

    def _plotonaxies(self, ax, **kwargs):
        for l in self.layers:
            l._plotonaxies(ax, **kwargs)
        super(CompoundPlot, self)._plotonaxies(ax, **kwargs)


class FilterColorPlotter(CompoundPlot):
    def __init__(self, distance=10, age=0, Av=None, reddenings=None,
                 collections=None, limits=None, **kwargs):
        self.lstars = CollectionLayer(collections, **kwargs)
        super(FilterColorPlotter, self).__init__(layers=[self.lstars], **kwargs)
        self.age = age
        self.distance = distance
        self._Av = Av
        self.reddenings = reddenings
        self.limits = {} if limits is None else limits

    @property
    def modulus(self): return 5*np.log10(self.distance, 10) - 5

    @property
    def Av(self):
        if self._Av is not None:
            return self._Av
        return self._get_reddening('B-V') * 3.1  # 1984A&AS...58..447D

    def update_limits(self, colors, limits):
        """
        Set axis limits for plots.

        Dictionary of limits is also available via `FilterColorPlotter.limits` property.
        Note, that limits can be changed after drawing ion returned axis object also.

        Existing limits, not listed in `colors`, remains unchanged.

        Parameters
        ----------
        colors : iterable of str
            color(s) for which limit is set
        limits : list of tuples or tuple
            pairs representing limits
            if limits is a single tuple, same limits are set for all `colors`

        Example
        -------

            plotter = JohnsonCousinsPlotter(U=u, B=b, V=v)
            plotter.update_limits('UBV', (22, 10))
            plotter.update_limits(['U-B', 'B-V'], (-2, 2))
        """
        try:
            iter(colors)
        except TypeError:
            colors = [colors]
        try:
            st = {c: (l[0], l[1]) for c,l in zip (colors, limits)}
        except TypeError:
            st = {c: (limits[0], limits[1]) for c in colors}
        self.limits.update(st)



    def _get_reddening(self, color):
        try:
            return self.reddenings[color]
        except (KeyError, TypeError):
            return 0

    def _plotonaxies(self, ax, **kwargs):
        kwargs.setdefault('xlim', self.limits.get(kwargs['xcolor']))
        kwargs.setdefault('ylim', self.limits.get(kwargs['ycolor']))
        super(FilterColorPlotter, self)._plotonaxies(ax, **kwargs)



class JohnsonCousinsPlotter(FilterColorPlotter):
    default_limits = {
        'U': (20,5), 'V': (20,5), 'B': (20,5), 'R': (20,5), 'I': (20,5),
        'U-B': (0,2.5), 'B-V': (0,2.5), 'V-I': (0,3.5), 'V-R': (0,2.5), 'R-I': (0,2.5),
        'U_e': (0, 0.3), 'B_e': (0, 0.3), 'V_e': (0, 0.3), 'R_e': (0, 0.3), 'I_e': (0, 0.3),
    }
    def __init__(self, U=None, B=None, V=None, R=None, I=None,
                 U_e=None, B_e=None, V_e=None, R_e=None, I_e=None,
                 distance=None, age=None, Av=None, reddenings=None, **kwargs):
        collections = {c: v for c,v in zip('UVBRI', [U, V, B, R, I]) if v is not None}
        for c in  'UVBRI':  # DiffPhots as arguments? extract mag and error
            dp = collections.get(c)
            if isinstance(dp, DiffPhot):
                collections[c] = dp.mag
                if dp.mag_e is not None:
                    collections[c+'_e'] = dp.mag_e
        errors = {c+'_e': v for c, v in zip('UVBRI', [U_e, V_e, B_e, R_e, I_e]) if v is not None}
        collections.update(errors)
        super(JohnsonCousinsPlotter, self).__init__(
            distance, age, Av, reddenings, collections, limits=dict(self.default_limits), **kwargs)


class MultiJohnsonCousinsPlotter(JohnsonCousinsPlotter):
    def __init__(self, U=None, B=None, V=None, R=None, I=None,
                 U_e=None, B_e=None, V_e=None, R_e=None, I_e=None,
                 colors=True, sizes=False,
                 **kwargs):
        U, B, V, R, I, U_e, B_e, V_e, R_e, I_e = (np.concatenate(t) for t in [U,B,V,R,I,U_e,B_e,V_e,R_e,I_e])
        subsets = np.concatenate([ np.full_like(ss, n) for n, ss in enumerate(U)])
        super(MultiJohnsonCousinsPlotter, self).__init__(
            U, B, V, R, I, U_e, B_e, V_e, R_e, I_e, **kwargs)
        self.lstars.subsets = subsets
        self.lstars.colors = colors
        self.lstars.sizes = sizes



class _TestPlotter:
    def color_mag_plot(ax, x, y, xcolor='B-V', ycolor='V', isochrone=None, sptypes=True, legend=None,
                       masks=[None], labels=None, colors=None, edgecolors=None, alphas=None, sizes=None,
                       xlim=None, ylim=None):
        def p(par, i): return None if par is None or par[i] is None else par[i]

        for i, mask in enumerate(masks):
            s = p(sizes, i)
            if s is None:
                s = 2*(20-y[mask])
            ax.scatter(x[mask], y[mask],
               label = p(labels,i), s=s, c=p(colors,i), edgecolors=p(edgecolors,i), alpha=p(alphas,i))
        # if sptypes:
        #     ax2 = sp.plot_twiny_spectral_types(ax, xcolor)
        # if isochrone:
        #     ax.plot(iB - iV, iV, 'r', linewidth=1, label='${:g}\mathrm{{Myr}}$ isochrone'.format(isochrone/1e6))

        #ax.plot(iso_sm.B_mag - iso_sm.V_mag, iso_sm.V_mag, 'rx') # solar mass points
        # for a in np.arange(7,9, 0.3):
        #     iso3 = MIST.isochrone(age=a, feh=0.001, AV=Av, distance=distance, maxm=18.25)
        #     ax.plot(iso3.B_mag - iso3.V_mag, iso3.V_mag, label='{:.3f}My isochrone'.format(10**a/1e6), linewidth=1, alpha=0.94)
        # iso4 = MIST.isochrone(age=7, feh=0.001)#, maxm=18.25)
        # ax.plot(iso4.B_mag - iso4.V_mag + E_B_V, iso4.V_mag+ distance_modulus + Av, label='10My isochrone', linewidth=1, alpha=0.94)
        # for a in np.arange(0,4.5, 0.5):
        #     iso3 = MIST.evtrack(2.0**a, minage=7, maxage=np.log10(20e6))
        #     ax.plot(iso3.B_mag - iso3.V_mag + E_B_V, iso3.V_mag+ distance_modulus + Av, label='{:.3f}Ms track'.format(2.0**a), linewidth=1, alpha=0.94)

        # ax.plot(iso['Bmag'] - iso['Vmag'] + E_B_V, iso['Vmag'] + distance_modulus + Av, 'r', alpha=1, linewidth=1,
        #        label='10My isochrone')
        ax.set_xlabel('$B-V$')
        ax.set_ylabel('$V$')
        ax.set_xlim(0, 2.5)
#        ax2.set_xlim(0, 2.5)
        ax.set_ylim(20,5)
        if isinstance(legend, bool):
            if legend:
                ax.legend()
        else:
            ax.legend(loc=legend)


if __name__ == "__main__":
    from astroquery.vizier import Vizier
    massey = Vizier.query_region('NGC 6871', radius='0d6m0s', catalog='J/ApJ/454/151/ngc6871')[0]
    V = massey['Vmag']
    B = massey['B-V'] + V
    U = massey['U-B'] + B
    plotter = JohnsonCousinsPlotter(U, B, V, distnace=1649, age=10e6, Av=1.364)
    ax = plotter.plot(xcolor='B-V', ycolor='V')
    plt.show()




# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os.path as path
from io import StringIO
import astwro.starlist as sl
import astwro.sampledata as data
from astwro.utils import tmpdir

def get_catalog(object):
    from astroquery.vizier import Vizier
    viz = Vizier( columns=['Star', '*'])
    star = object.replace('_', ' ') + '*'
    try:
        return viz.query_constraints(catalog='II/183A/table2',Star=star)[0]
    except IndexError:
        return None

def test_read_write_ds9():
    s1 = sl.read_dao_file(data.ap_file())
    d = tmpdir()
    f2 = path.join(d.path, 'i.reg')
    sl.write_ds9_regions(s1, f2)
    s2 = sl.read_ds9_regions(f2)
    assert s2[['id', 'x', 'y']].equals(s1[['id', 'x', 'y']])
    assert not s2.auto_id.any()

def test_read_noid_reg():
    reg = u"""
# Region file format: DS9 version 4.1
global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
image
circle(869.40377,745.33678,13.888889)
circle(1225.6722,742.09608,13.888889)
circle(753.77706,465.33857,13.888889)
circle(1034.8725,499.95079,13.888889)
circle(1194.5182,211.78505,13.888889)
    """
    regstrm = StringIO(reg)
    s = sl.read_ds9_regions(regstrm)
    assert s.id[1] == 1
    assert s.x[3] == 753.77706
    assert s.auto_id.all()


def test_read_mixedid_reg():
    reg = u"""
# Region file format: DS9 version 4.1
global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
image
circle(869.40377,745.33678,13.888889)
circle(1225.6722,742.09608,13.888889)
circle(753.77706,465.33857,13.888889)  #  id=160
circle(1034.8725,499.95079,13.888889)
circle(1194.5182,211.78505,13.888889)
    """
    regstrm = StringIO(reg)
    s = sl.read_ds9_regions(regstrm)
    assert s.x[161] == 869.40377
    assert s.y[160] == 465.33857
    assert s.auto_id.any()
    assert not s.auto_id.all()


def test_read_wcs_reg():
    reg = u"""
# Region file format: DS9 version 4.1
global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
fk5
circle(21:45:00.7278,+65:45:56.751,2.275") # color=#38ACFB text={1} id=1
circle(21:45:33.1351,+65:50:07.545,2.275") # color=#38ACFB text={7} 
circle(21:45:41.7344,+65:45:45.371,2.275") # color=#38ACFB text={8} id=8
circle(21:45:46.9471,+65:45:43.794,2.275") # color=#38ACFB text={9} id=9
    """
    regstrm = StringIO(reg)
    s = sl.read_ds9_regions(regstrm)
    assert s.ra[8] == '21:45:41.7344'
    assert s.dec[1] == '+65:45:56.751'
    assert s.auto_id.any()
    assert not s.auto_id.all()

def test_write_wizir_catalog():
    d = tmpdir()
    fpath = path.join(d.path, 'mark_a.reg')
    std_catalog = get_catalog('MARK_A')
    sl.write_ds9_regions(std_catalog, fpath, WCS=True)
    cat = sl.read_ds9_regions(fpath)
    assert len(cat) == len(std_catalog)





# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os.path as path
import astwro.starlist as sl
import astwro.sampledata as data
from astwro.utils import tmpdir


def check_starlist_daotype(s, daotype):
    assert s.DAO_type == daotype
    assert s.stars_number() > 5
    assert s.columns[0] == 'id'
    assert (s.index == s.id).all()

def check_readwritereadqueals(f1):
    s1 = sl.read_dao_file(f1)
    d = tmpdir()
    f2 = path.join(d.path, 'tmp' + s1.DAO_type.extension)
    sl.write_dao_file(s1, f2)
    s2 = sl.read_dao_file(f2)
    assert s1.equals(s2)


def test_read_coo():
    s = sl.read_dao_file(data.coo_file())
    check_starlist_daotype(s, sl.DAO.COO_FILE)

def test_read_lst():
    s = sl.read_dao_file(data.lst_file())
    check_starlist_daotype(s, sl.DAO.LST_FILE)

def test_read_nei():
    s = sl.read_dao_file(data.nei_file())
    check_starlist_daotype(s, sl.DAO.NEI_FILE)

def test_read_ap():
    s = sl.read_dao_file(data.ap_file())
    check_starlist_daotype(s, sl.DAO.AP_FILE)

def test_read_als():
    s = sl.read_dao_file(data.als_file())
    check_starlist_daotype(s, sl.DAO.ALS_FILE)



def test_write_coo():
    check_readwritereadqueals(data.coo_file())

def test_write_lst():
    check_readwritereadqueals(data.lst_file())

def test_write_nei():
    check_readwritereadqueals(data.nei_file())

def test_write_ap():
    check_readwritereadqueals(data.ap_file())

def test_write_als():
    check_readwritereadqueals(data.als_file())

def test_convert_ap_to_als():
    s1 = sl.read_dao_file(data.als_file())
    d = tmpdir()
    f2 = path.join(d.path, 'tmp' + '.als')
    sl.write_dao_file(s1, f2, sl.DAO.ALS_FILE)
    s2 = sl.read_dao_file(f2)
    cols_to_compare = ['id', 'x', 'y', 'mag', 'sky']
    assert s1.round(4)[cols_to_compare].equals(s2.round(4)[cols_to_compare])





#! /usr/bin/env python
# coding=utf-8
""" Grep-like tool for FITS headers

    Call commandline: ``grepfitshdr --help`` for parameters info.
"""
from __future__ import print_function, division

import astropy.io.fits as pyfits
import re
from os.path import isfile
from sys import exit
from sys import stdout
from glob import glob

import astwro.tools.__commons as commons


def headers(filenames, hdus):
    for fname in filenames:
        if isfile(fname):
            fits = pyfits.open(fname)
            for hdu in hdus:
                try:
                    fits[hdu].verify('silentfix')
                    h = fits[hdu].header
                    yield h, fname, hdu
                except IndexError:
                    print('[WARN] HDU {} non exist in file {}'.format(hdu, fname))
            fits.close()


def printmatch(output, filename, hdu, line, withfile, fileonly):
    if not fileonly:
        if withfile:
            print (filename + ' hdu:{}: '.format(hdu), end='', file=output)
        print (line, file=output)

def iter_fields(hdr, onlyvalues=False, fields=None):
    """ splits header into lines
        if onlyvalues does not return field names
        if fields returns only specified fields (forces onlyvalues)"""
    if fields:
        for key in hdr:
            if key in fields:
                yield hdr[key]
    elif onlyvalues:
        for val in hdr.values():
            if val:
                yield str(val)
    else:
        for line in repr(hdr).strip().splitlines():
            yield line


def grep(pattern, filenames, output=stdout, invert=False, withfile=False, fileonly=False,
         fields=None, onlyvalues=False, ignorecase=True, hdu=None):
    if fields is not None and fields[0] == '*':
        fields = None
        onlyvalues = True
    if hdu is None:
        hdu = [0]
    if isinstance(filenames, str):
        filenames = glob(filenames)
    regexp = re.compile(pattern, flags=re.IGNORECASE if ignorecase else  0)
    globmatched = 0
    for h, f, n in headers(filenames, hdus=hdu):
        matched = scanned = 0
        # rep = repr(h).strip()
        # for line in rep.splitlines():
        for line in iter_fields(h, onlyvalues=onlyvalues, fields=fields):
            match = regexp.search(str(line))
            if invert:
                match = not match
            scanned += 1
            if match:
                matched += 1
                printmatch(output, f, n, line, withfile, fileonly)
                if fileonly and not invert:
                    break
        globmatched += matched
        if fileonly:
            if (not invert and matched > 0) or (invert and matched == scanned):
                print (f, file=output)
    return globmatched

def __do(arg):
    if isinstance(arg.hdu, str):
        arg.hdu = [int(n) for n in arg.hdu.split(',')]
    return grep(arg.pattern, arg.file, invert=arg.v, withfile=arg.H,
                fileonly=arg.l, fields=arg.f, ignorecase=arg.i, hdu=arg.hdu)


def __arg_parser():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        fromfile_prefix_chars='@',
        epilog='exit code:\n'
               '  0 if any header matched pattern\n'
               '  1 if no match found\n\n' + commons.version_string(),
        description='grep-like utility for fits (main) headers\n '
                    'until -f specified, searches in keys, values and comments')
    parser.add_argument('pattern', type=str,
                        help='reg-exp, use single dot . to dump all header fields')
    parser.add_argument('file', type=str, nargs='+',
                        help='FITS file(s), catalog file containing file names prefixed by @ can be provided')
    parser.add_argument('-i', action='store_true',
                        help='ignore case')
    parser.add_argument('-v', action='store_true',
                        help='invert match')
    parser.add_argument('-H', action='store_true',
                        help='add filename to each found line')
    parser.add_argument('-l', action='store_true',
                        help='print filenames with matches only')
    parser.add_argument('-f', action='append', metavar='FIELD',
                        help='matches only specified FIELD\'s value; can be provided multiple '
                             'times to match several fields; -f* limits search to values but searches '
                             'in all fields')
    parser.add_argument('-n', '--hdu', metavar='x,y,z,...',
                        help='HDU(s) to scan, comma-separated-list; default: 0')
    return parser


def main(pattern, file, **kwargs):
    """Entry point for python script calls. Parameters identical to command line"""
    # Extract default arguments from command line parser and apply kwargs parameters
    args = commons.bunch_kwargs(__arg_parser(), positional=[pattern, file], **kwargs)
    # call main routine - common form command line and python calls
    return __do(args)


def info():
    """Prints commandline help message"""
    commons.info(__arg_parser())


def commandline_entry():
    # Entry point for command line
    __args = __arg_parser().parse_args()  # parse command line arguments
    n = __do(__args)  # call main routine - common form command line and python calls
    return  0 if n > 0 else 1

if __name__ == '__main__':
    code = commandline_entry()
    exit(code)



from __future__ import print_function, division


def bunch_kwargs(parser, positional=[], **kwargs):
    """Makes object from arguments dict. used to have common interface
    for arguments passed to main() from command line or calling script."""
#    return namedtuple('ArgsBunch', kwargs.keys())(*kwargs.values())
#    parser.argument_default='==SUP=='
#    parser.set_defaults(**kwargs)
    arg = parser.parse_args(positional)   # create arguments namespace with defaults
    vars(arg).update(kwargs) # update with provided kwargs
    return arg


def version_string():
    from ._version import __version__
    return 'AstWro tools v. {} [github.com/majkelx/astwro]'.format(__version__)


def main_info():
    return 'For use from python call main function foo(**kwargs), with command line arguments as named arguments'


def info(parser):
    parser.print_help()
    print (main_info())




#! /usr/bin/env python
# coding=utf-8
""" Find best PSF stars using GA to minimize mean error.

    .. seealso:: :ref:`gapick`
"""
from __future__ import absolute_import, division, print_function
__metaclass__ = type

# For now:
# TODO: --neigbours option: once for generation neightbour removal
# TODO: --timestamp: add current timestamp to directory name (+)
# TODO: conf option for
# TODO: warnings about: daophot.opt, allstar.opt
# TODO: single parameter --aperture, instead of photo.opt
# TODO: write down text based statistics from logbook.pkl
# TODO: some docs, and link to them in --help
# TODO: allstar.opt missing in result dir!
# For later, on request:
# TODO: Option for provide own PSF stars set, used for comaprision, including in candidates, put in initial population
# TODO: Checkpoints - continue calculation



import os
import logging
import random
import pickle
import time
from datetime import timedelta
from copy import deepcopy

try:
    from itertools import izip_longest as zip_longest
except ImportError:
    from itertools import zip_longest

import numpy
from bitarray import bitarray
from scipy.stats import sigmaclip
from deap import base
from deap import creator
from deap import tools

import astwro.tools.__commons as commons
import astwro.starlist as sl
import astwro.pydaophot as dao
import astwro.tools
import astwro.utils as utils
from astwro.phot import PhotError

_time_format = '%a %H:%M:%S'


# Definitions for genetic algorithm:
# 'individual' - subset of candidates competing with other subsets to be the best in fitness
# 'genome' - the individual's definition with subset of candidates stars, is represented by binary string
#     of length equal to number of candidate stars. '1' means that star is in subset.
# 'fitness' - minimized function on individual, mean of errors form allstar if individual's stars are
#     taken as PSF stars
# 'population' - set of all individuals in some iteration

# 2.1 Definition of routines used by algorithms: initializations, scoring
def select_stars(starlist, genome):
    # type: (sl.StarList, bitarray) -> sl.StarList
    # Select stars present in genome
    return starlist[genome.tolist()]


def random_genome(ind_class, len, prob_limits):
    # type: (type, int, [float, float]) -> bitarray
    # Create random genome of length `len` in which probability of '1' on any position is from `prob_limits` range.
    prob = random.uniform(prob_limits[0], prob_limits[1])
    return ind_class([random.random() <= prob for _ in range(len)])


def clone_individual(individual):
    # type: (bitarray) -> bitarray
    # Make a deepcopy-clone - deepcopy genome bitarray and associated fitness
    n = deepcopy(individual)
    n.fitness = deepcopy(individual.fitness)
    return n


def calc_spectrum(pop):
    # calculates 'spectrogram' of population
    # which is a statistic of stars occurrences in population individuals
    spec = numpy.zeros(len(pop[0]))
    for ind in pop:
        spec += ind.tolist()
    return spec


# === astwro version <= 0.7.0
# def fitness_for_als(als):
#     # type: (sl.StarList) -> (float,)
#     # Calucalates fitness from allstar result
#     clipped = sigmaclip(als.chi)[0]
#     return numpy.sqrt((clipped * clipped).sum()),  # fitness is tuple (val,)

# === astwro version > 0.7.0
# def fitness_for_als(als):
#     # type: (sl.StarList) -> (float,)
#     # Calucalates fitness from allstar result
#     w = 100*(-als.mag/5)
#
#     return (als.chi*w).sum()/w.sum(),  # fitness is tuple (val,)

# === astwro version > 0.7.2
def fitness_for_als(als):
    # type: (sl.StarList) -> (float,)
    # Calucalates fitness from allstar result
    w = 100*(-als.mag/5)

    return (als.mag_err*w).sum()/w.sum(),  # fitness is tuple (val,)


def eval_population(population, candidates, workers, show_progress, fine_tune):
    if fine_tune:
        return eval_population_fine_psf(population, candidates, workers, show_progress)
    else:
        return eval_population_simple(population, candidates, workers, show_progress)


def eval_population_fine_psf(population, candidates, workers, show_progress):
    # type: (list(bitarray), sl.StarList, list(dict), bool) -> list
    # Evaluates fitness for all individual in population.

    # This version uses sofisticated process from daophot_bialkow
    # Uses daophots and allstars processes form `workers` running them in parallel.
    # :return: list fitnesses (1-element couples as `deap` lib likes)

    progress = None
    if show_progress:
        progress = utils.progressbar(total=len(population), step=len(workers))
        progress.print_progress(0)
    fitnesses = []
    f_max = None
    # https://docs.python.org/3/library/itertools.html#itertools-recipes grouper()
    # grouping population into chunks of size `parallel`
    for chunk in zip_longest(*([iter(population)] * len(workers))):
        active = [g is not None for g in chunk]  # all active, on errors some could be deactivated
        # PSF
        for individual, worker in zip(chunk, workers):
            if individual:
                pdf_s = select_stars(candidates, individual)
                worker['daophot'].write_starlist(pdf_s, 'i.lst')
                worker['daophot'].PSf(psf_stars='i.lst')  # add to queue only - batch mode
                worker['daophot'].run(wait=False)  # parallel start all workers without waiting
        # wait for PSF and start ALLSTAR nei
        for i, worker in enumerate(workers):
            if active[i]:
                worker['daophot'].wait_for_results()  # now wait before using results
                active[i] = worker['daophot'].PSf_result.converged  # PSF is not always successful
                if active[i]:
                    worker['allstar'].ALlstar(stars='i.nei')  # enqueue calculation
                    worker['allstar'].run(wait=False)  # asynchronous / parallel
        # second PSF
        for i, worker in enumerate(workers):
            if active[i]:
                worker['allstar'].wait_for_results()
                if worker['allstar'].ALlstars_result.success:
                    worker['daophot'].SUbstar(subtract='i.als', leave_in='i.lst')
                    worker['daophot'].run()  # quick run
                    worker['daophot'].ATtach('is')
                    worker['daophot'].PSf(photometry='i.als', psf_stars='i.lst')
                    worker['daophot'].run(wait=False)
                else:
                    active[i] = False
        # second allstar
        for i, worker in enumerate(workers):
            if active[i]:
                worker['daophot'].wait_for_results()
                if worker['daophot'].PSf_result.success:
                    worker['allstar'].ALlstar(stars='i.nei')  # enqueue calculation
                    worker['allstar'].run(wait=False)  # asynchronous / parallel
                else:
                    active[i] = False
        # third PSF
        for i, worker in enumerate(workers):
            if active[i]:
                worker['allstar'].wait_for_results()
                if worker['allstar'].ALlstars_result.success:
                    worker['daophot'].SUbstar(subtract='i.als', leave_in='i.lst')
                    worker['daophot'].run()  # quick run
                    worker['daophot'].ATtach('is')
                    worker['daophot'].PSf(photometry='i.als', psf_stars='i.lst')
                    worker['daophot'].run(wait=False)
                else:
                    active[i] = False
        # final allstar
        for i, worker in enumerate(workers):
            if active[i]:
                worker['daophot'].wait_for_results()
                if worker['daophot'].PSf_result.success:
                    worker['allstar'].ALlstar(stars='als.ap')  # enqueue calculation
                    worker['allstar'].run(wait=False)  # asynchronous / parallel
                else:
                    active[i] = False

        # wait for allstar for workers and process results
        for i, worker in enumerate(workers):
            if active[i]:
                worker['allstar'].wait_for_results()
                all_s = worker['allstar'].ALlstars_result.als_stars
                f = fitness_for_als(all_s)
                fitnesses.append(f)  # fitness is tuple (val,)
                f_max = f if f_max is None or f_max[0] < f[0] else f_max
            #                f_max = f[0] if f_max is None else max(f_max, f[0])
            else:
                fitnesses.append(None)
        # cut fitnesses if longer than population
        # (finesses mod parallel == 0, some None at the end can appear if population mod parallel != 0)
        fitnesses = fitnesses[:len(population)]
        # fill gaps in fitnesses with maximum of rest of population
        for i, f in enumerate(fitnesses):
            if f is None:
                fitnesses[i] = f_max
        if progress:
            progress.print_progress()

    return fitnesses


def eval_population_simple(population, candidates, workers, show_progress):
    # type: (list(bitarray), sl.StarList, list(dict), bool) -> list
    # Evaluates fitness for all individual in population.
    # Uses daophots and allstars processes form `workers` running them in parallel.
    # :return: list fitnesses (1-element couples as `deap` lib likes)

    progress = None
    if show_progress:
        progress = utils.progressbar(total=len(population), step=len(workers))
        progress.print_progress(0)
    fitnesses = []
    f_max = None
    # https://docs.python.org/3/library/itertools.html#itertools-recipes grouper()
    # grouping population into chunks of size `parallel`
    for chunk in zip_longest(*([iter(population)] * len(workers))):
        active = [g is not None for g in chunk]  # all active, on errors some could be deactivated
        # PSF
        for individual, worker in zip(chunk, workers):
            if individual:
                pdf_s = select_stars(candidates, individual)
                worker['daophot'].PSf(psf_stars=pdf_s)  # add to queue only - batch mode
                worker['daophot'].run(wait=False)  # parallel start all workers without waiting
        # wait for PSF and start ALLSTAR
        for i, worker in enumerate(workers):
            if active[i]:
                worker['daophot'].wait_for_results()  # now wait before using results
                active[i] = worker['daophot'].PSf_result.converged  # PSF is not always successful
                if active[i]:
                    worker['allstar'].ALlstar(stars='als.ap')  # enqueue calculation
                    worker['allstar'].run(wait=False)  # asynchronous / parallel
        # wait for allstar for workers and process results
        for i, worker in enumerate(workers):
            if active[i]:
                worker['allstar'].wait_for_results()
                all_s = worker['allstar'].ALlstars_result.als_stars
                f = fitness_for_als(all_s)
                fitnesses.append(f)  # fitness is tuple (val,)
                f_max = f if f_max is None or f_max[0] < f[0] else f_max
            #                f_max = max(f_max, f)
            else:
                fitnesses.append(None)
        # cut fitnesses if longer than population
        # (finesses mod parallel == 0, some None at the end can appear if population mod parallel != 0)
        fitnesses = fitnesses[:len(population)]
        # fill gaps in fitnesses by maximum of rest of population
        for i, f in enumerate(fitnesses):
            if f is None:
                fitnesses[i] = f_max
        if progress:
            progress.print_progress()

    return fitnesses


def _prepare_output_dir(outdir, overwrite, srcdir, arg):
    # type: (str, bool, str, object) -> (utils.CycleFile, utils.CycleFile, utils.CycleFile, str)
    # Prepare output directory for results
    if outdir:
        from shutil import copytree, rmtree

        outdir = os.path.abspath(os.path.expanduser(outdir))
        if os.path.exists(outdir):
            if overwrite:
                rmtree(outdir)
            else:
                logging.error('--out-dir:{} already exists and no --overwrite requested.'.format(outdir))
                raise Exception('Output directory {} exists.'.format(outdir))
        copytree(srcdir, outdir)
        logging.info('Results dir created: {}'.format(outdir))
        # todo allstar.opt missing in result dir!
        basename = 'gen'
        lst_file = utils.cyclefile(outdir, basename, '.lst')
        reg_file = utils.cyclefile(outdir, basename, '.reg')
        gen_file = utils.cyclefile(outdir, basename, '.gen')
        # write down parameters
        with open(os.path.join(outdir, 'about.txt'), 'w') as f:
            print("astwro pydaophot/tools version: {}/{}\n".format(dao.__version__, astwro.tools.__version__), file=f)
            for k, v in arg.__dict__.items():
                print('{}\t= {}'.format(k, v), file=f)
        return lst_file, reg_file, gen_file, outdir
    else:
        return None, None, None, None


def _plot_err(pe, pe_psf, psf, fig_file):
    # type: (PhotError) -> None
    try:
        import matplotlib
        matplotlib.use('pdf')
        import matplotlib.pyplot as plt
    except ImportError:
        logging.warning("'matplotlib' cannot be imported. No control diagrams will be generated")
    except:
        logging.warning('matplotlib PDF output cannot be initialized. No control diagrams will be generated')
        return
    else:
        try:
            plt.rc('text', usetex=True)
        except:
            pass

        ax = plt.subplot()
        ls = numpy.linspace(pe.mag.min(), pe.mag.max())
        ax.plot(pe.mag[~pe.outlayers_mask], pe.err[~pe.outlayers_mask], '.', alpha=0.7, ms=2,
                label='training set')
        ax.errorbar(pe.mag_means, pe.err_means, yerr=pe.err_means_stderr * 1.0, fmt='D', markerfacecolor='none', alpha=0.7, lw=1,
                    label='error bins')
        ax.plot(ls, pe.evaluate_fit(ls), lw=1,
                label='error fit')
        ax.plot(ls, pe.evaluate_fit_plus_sigmas(ls),  '--', lw=1,
                label='error$+{:.1f}\\sigma$ fit (training limit)'.format(pe.fitplussigma))
        ax.plot(ls, pe_psf.evaluate_fit_plus_sigmas(ls),  '--', lw=1,
                label='error$+{:.1f}\\sigma$ fit (PSF limit)'.format(pe_psf.fitplussigma))
        ax.plot(pe.mag[pe.outlayers_mask], pe.err[pe.outlayers_mask], '.', alpha=0.7, ms=3,
                label='outlayers')
        ax.plot(psf.mag, psf.mag_err, '.', alpha=0.7, ms=3,
                label='PSF candidates')
        ax.set_xlabel('mag'); ax.set_ylabel('mag error')
        ax.legend()
        ax.figure.savefig(fig_file, bbox_inches='tight')
        ax.set_ylim(0,0.2)
        ax.figure.savefig('{}_zoomed{}'.format(*os.path.splitext(fig_file)), bbox_inches='tight')
        logging.info('Control plot mag/err {} generated'.format(fig_file))


def __do(arg):
    # Main routine, common for command line, and python scripts call
    # :type arg: Namespace

    start_time = time.time()

    # Configure progress logging
    clogger = logging.getLogger('clean logger')  # create another logger for stats without prefixes (clean)
    chandler = logging.StreamHandler()
    chandler.setFormatter(logging.Formatter('%(message)s'))
    clogger.propagate = False
    clogger.addHandler(chandler)

    check_arguments(arg)

    # image_file
    if arg.image is None:
        from astwro.sampledata import fits_image
        arg.image = fits_image()  # sample image

        logging.warning('No image file argument provided (-h for help), using demonstration image: %s', arg.image)

    # expand patches of file parameters
    arg.all_stars_file = dao.Daophot.expand_path(arg.all_stars_file)
    arg.psf_stars_file = dao.Daophot.expand_path(arg.psf_stars_file)
    arg.photo_opt = dao.Daophot.expand_path(arg.photo_opt)
    arg.daophot_opt = dao.Daophot.expand_path(arg.daophot_opt)

    # get single daophot and ATtach file
    dp = dao.Daophot(image=arg.image, daophotopt=arg.daophot_opt)
    al = dao.Allstar(image=arg.image, dir=dp.dir)

    # all stars file
    if not arg.all_stars_file:
        logging.warning('No all-stars-file provided, Stars will be found by daophot FIND (frames av/sum: %d/%d)',
                        arg.frames_av, arg.frames_sum)
        find = dp.FInd(arg.frames_av, arg.frames_sum)
        logging.info('{} stars found by FIND, sky estimation: {}, err/dev: {}/{}'.format(find.stars, find.sky, find.err,
                                                                                         find.skydev))
        arg.all_stars_file = find.starlist_file

    # photometry
    if not arg.photo_opt:  # no file no problem, but recreate default radius if not provided
        if arg.photo_is == 0:
            arg.photo_is = 35
        if arg.photo_os == 0:
            arg.photo_os = 50
        if not arg.photo_ap:
            arg.photo_ap = [8]
    logging.info('Performing preliminary {} pixels aperture  photometry'.format(arg.photo_ap[0]))
    photometry = dp.PHotometry(photoopt=arg.photo_opt,
                               IS=arg.photo_is,
                               OS=arg.photo_os,
                               apertures=arg.photo_ap,
                               stars=arg.all_stars_file)

    # pick PFS stars candidates
    PICKcandidates = not arg.psf_stars_file
    if PICKcandidates:
        pick = dp.PIck(arg.stars_to_pick, arg.faintest_to_pick)
        arg.psf_stars_file = pick.picked_stars_file

    # psf (for errors collection)
    PSf_result = dp.PSf(psf_stars=arg.psf_stars_file)

    # all stars (filtering by photometry errors and magnitudes)
    stars = photometry.photometry_starlist
    count0 = stars.shape[0]
    stars = stars[stars.mag < arg.max_ph_mag]
    count1 = stars.shape[0]
    stars = stars[stars.mag_err < arg.max_ph_err]
    count2 = stars.shape[0]
    logging.info('{} (of {}) stars left after filtering against {:.2f} photometry magnitude threshold, '
                 .format(count1, count0, arg.max_ph_mag))
    logging.info('{}  left after {:.2f} aperture photometry error threshold'
                 .format(count2, arg.max_ph_err))

    logging.info('Performing preliminary PSF profile photometry')
    ALlstar_result = al.ALlstar(stars=stars)
    logging.info('{} stars converged in PSF profile photometry'
                 .format(ALlstar_result.stars_no[0]))
    logging.info('Fitting PSF profile photometry errors')
    pe = PhotError(ALlstar_result.als_stars.mag, ALlstar_result.als_stars.mag_err, fitplussigma=arg.clip_fit_sigmas)
    logging.info('{} stars left, after excluding profile photometry error {} sigma outliers'
                 .format((~pe.outlayers_mask).sum(), arg.clip_fit_sigmas))
    stars = ALlstar_result.als_stars[~pe.outlayers_mask]

    # for the sake of optimisation we write starlist to file to avoid multiple serialization of this list
    # (instead of more obvious providing starlist as the allstar argument)
    dp.write_starlist(stars, 'als.ap', sl.DAO.AP_FILE)

    # candidates (filter out big psf errors)
    pe2 = None
    if PICKcandidates: # do it again
        if arg.clip_fit_sigmas > arg.clip_psf_fit_sigmas: # tighter filtering for PSF candids
            logging.info('Excluding {} sigma outlayers for PICK'.format(arg.clip_psf_fit_sigmas))
            pe2 = PhotError(ALlstar_result.als_stars.mag, ALlstar_result.als_stars.mag_err,
                           fitplussigma=arg.clip_psf_fit_sigmas)
            cands_for_cands = ALlstar_result.als_stars[~pe2.outlayers_mask]
        else:
            cands_for_cands = stars
        pick = dp.PIck(arg.stars_to_pick, arg.faintest_to_pick, photometry=cands_for_cands)
        logging.info('{} PSF stars candidates selected by PICK'.format(pick.stars))
        arg.psf_stars_file = pick.picked_stars_file
        PSf_result = dp.PSf(psf_stars=arg.psf_stars_file) # collect PSF errors
    candidates = dp.read_starlist(arg.psf_stars_file, add_psf_errors=True)
    org_cand_no = candidates.stars_number()
    err = PSf_result.errors
    averr = err.psf_err.mean()
    err = err[(err.psf_err < arg.max_psf_err_mult * averr) & (
                err.flag == ' ')]  # filter out big errors and * or ? marked stars
    candidates = candidates.loc[err.index]

    # candidates = candidates[candidates.psf_err < arg.max_psf_err]  # old filter, new above
    logging.info(
        "{} good candidates ({} rejected: * or ? or psf error exceeded max-psf-err-mult*averge-error "
        "= {:.2f}*{:.2f} = {:.2f})"
            .format(
            candidates.stars_number(),
            org_cand_no - candidates.stars_number(),
            arg.max_psf_err_mult,
            averr,
            arg.max_psf_err_mult * averr)
    )
    if arg.include_psf:
        logging.info('PSF star candidates stays in training set (--include-psf)')
    else:
        overlapping = stars.index.isin(candidates.index)
        stars = stars[~overlapping]
        logging.info('{} stars left in training set after exclusion PSF star candidates (no --include-psf)'
                     .format(stars.stars_number()))

    if candidates.stars_number() < 15:
        logging.error("Number of candidates lass than 15. GA needs more. Sorry...")

    if stars.stars_number() < 15:
        logging.error("Number stars in training set lass than 15. GA needs more. Sorry...")

    # Prepare output directory for results
    lst_file, reg_file, gen_file, result_dir = _prepare_output_dir(arg.out_dir, arg.overwrite, str(dp.dir), arg)
    _plot_err(pe, pe2, candidates, os.path.join(arg.out_dir, 'prel_psf_filtering.pdf'))

    #  From all candidates find best subset, where best means minimizing mean of errors form allstar
    #       using genetic algorithm
    #  For details about setting up genetic algorithm with DEAP visit:
    #       http://deap.gel.ulaval.ca/doc/default/examples/ga_onemax.html

    creator.create("FitnessMax", base.Fitness, weights=(-1.0,))
    creator.create("Individual", bitarray, fitness=creator.FitnessMax)

    toolbox = base.Toolbox()
    # Structure initializes
    toolbox.register("individual", random_genome, creator.Individual, candidates.stars_number(), arg.ga_init_prob)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register('clone', clone_individual)

    # The Genetic Operators

    # set min_stars to all_cand_no*ga_init_prob/2
    toolbox.register('mate', tools.cxTwoPoint)
    toolbox.register('mutate', tools.mutFlipBit, indpb=arg.ga_mut_str)
    toolbox.register('select', tools.selTournament, tournsize=3)

    # setup stats
    stats_fits = tools.Statistics(key=lambda ind: ind.fitness.values)
    stats_star = tools.Statistics(key=sum)  # number of stars is an sum of bitarray: [001101010001] has 5 stars
    stats = tools.MultiStatistics(fitness=stats_fits, size=stats_star)
    stats.register('avg', numpy.mean)
    stats.register('std', numpy.std)
    stats.register('min', numpy.min)
    stats.register('max', numpy.max)

    # Initiate workers. Each worker has Daophot and Allstar objects sharing runner directory,
    # working in batch mode.
    workers = []
    workers_logger = logging.getLogger('worker')
    workers_logger.setLevel('ERROR')  # prevent workers flood output with logrecords
    for i in range(arg.parallel):
        d = dp.clone()  # clone previously used daophot
        d.batch_mode = True
        a = dao.Allstar(dir=d.dir, image=d.image, batch=True, options={'MA': 100})
        d.logger = workers_logger
        a.logger = workers_logger
        workers.append({'daophot': d, 'allstar': a})

    # Setup initial population, HoF and logbook and  or load it from checkpoint when continuing previous calculation
    start_gen = 0

    #    if arg.checkpoint:   ## Not implemented
    if False:
        with open(os.path.expanduser(arg.checkpoint), "rb") as f:
            checkpoint = pickle.load(f)
        pop = checkpoint['population']
        start_gen = checkpoint['generation']
        hof = checkpoint['halloffame']
        logbook = checkpoint['logbook']
        logging.info('Restoring genetic algorithm on {} of {} generations'.format(start_gen, arg.ga_max_iter))
    else:
        hof = tools.HallOfFame(maxsize=10)

        logbook = tools.Logbook()
        logbook.header = 'gen', 'fitness', 'size'
        logbook.chapters['fitness'].header = 'min', 'avg', 'max', 'std'
        logbook.chapters['size'].header = 'min', 'avg', 'max'

        pop = toolbox.population(n=arg.ga_pop)
        logging.info('Starting genetic algorithm for {} generations at {}'.format(
            arg.ga_max_iter,
            time.strftime(_time_format, time.localtime())
        ))
        logging.info('{} parallel threads, ETA will be calculated after generation 1'.format(arg.parallel))

    # Calculate fitnesses of initial population
    fitnesses = eval_population(pop, candidates, workers, show_progress=not arg.no_progress, fine_tune=arg.fine)
    for ind, fit in zip(pop, fitnesses):
        ind.fitness.values = fit

    record = stats.compile(pop)
    logbook.record(gen=0, spectrum=calc_spectrum(pop), **record)
    clogger.info('{}\t ETA: [... to be determined]'.format(logbook.stream))

    evolution_start_time = time.time()

    # Begin the evolution
    for g in range(start_gen + 1, arg.ga_max_iter):
        #  New Generation
        #  select the next generation individuals
        offspring = toolbox.select(pop, len(pop))
        # Clone the selected individuals
        offspring = list(map(toolbox.clone, offspring))
        # Apply crossover and mutation on the offspring
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < arg.ga_cross_prob:
                toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values
        for mutant in offspring:
            if random.random() < arg.ga_mut_prob:
                toolbox.mutate(mutant)
                del mutant.fitness.values

        # calculate fitnesses of new individuals
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = eval_population(invalid_ind, candidates, workers, show_progress=not arg.no_progress,
                                    fine_tune=arg.fine)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
        # New population from offspring
        pop[:] = offspring

        # Stats
        # hof.update(pop)  # not implemented yet, __deapcopy__ of the Individual should work first
        ETA = time.strftime(_time_format, time.localtime(
            evolution_start_time + (time.time() - evolution_start_time) * arg.ga_max_iter / g))
        record = stats.compile(pop)
        logbook.record(gen=g, spectrum=calc_spectrum(pop), **record)
        clogger.info('{}\t ETA: {}'.format(logbook.stream, ETA))

        # For every generation create lst file and ds9 reg file of best and point symlinks to last generation
        if result_dir:
            best_ind = tools.selBest(pop, 1)[0]
            best_stars = select_stars(candidates, best_ind)
            lst_file.next_file(g)
            reg_file.next_file(g)
            gen_file.next_file(g)
            sl.write_dao_file(best_stars, lst_file.file, sl.DAO.LST_FILE)
            sl.write_ds9_regions(best_stars, reg_file.file)
            for ind in pop:
                gen_file.file.write(ind.to01() + '\n')
            with open(os.path.join(result_dir, 'logbook.pkl'), 'wb') as f:
                pickle.dump(logbook, f)
            checkpoint = dict(population=pop, generation=g, halloffame=hof, logbook=logbook)
            with open(os.path.join(result_dir, 'checkpoint.chk'), 'wb') as f:
                pickle.dump(checkpoint, f)

        # end of evolution loop

    if result_dir:
        lst_file.close()
        reg_file.close()
        gen_file.close()

    logging.info('Successful evolution finished at {} (elapsed time: {:s})'.format(
        time.strftime(_time_format, time.localtime()),
        str(timedelta(seconds=time.time() - start_time))
    ))

    best_ind = tools.selBest(pop, 1)[0]
    logging.info('Best individual is {}, {}'.format(best_ind, best_ind.fitness.values))

    best_stars = select_stars(candidates, best_ind)
    return best_stars


__arg_parser_singlethon = None


def __arg_parser():
    global __arg_parser_singlethon
    if __arg_parser_singlethon:
        return __arg_parser_singlethon
    import argparse
    parser = argparse.ArgumentParser(
        description='Find best PSF stars using genetic algorithm running the daophot/allstar'
                    ' workflow for number of a PSF stars sets.'
                    ' Minimized function'
                    ' is the mean of allstar\'s chi statistic, weighted by flux, calculated on '
                    ' training subset of of all stars. Results'
                    ' are stored in --dir directory if provided. List of star ids'
                    ' of the best set found, is printed to stdout (until suppressed by -no_stdout)',
        epilog=commons.version_string(),
    )
    parser.add_argument('image', default=None, nargs='?',
                        help='FITS image file (obligatory if no --demo)')
    g_dao = parser.add_argument_group('DAOPHOT/ALLSTARS related parameters',
                                      'Those parameters are used by DAOPHOT/ALLSTARS workflow. gapick optimizes PSF '
                                      'fitting errors of that workflow applied to training set stars as a function'
                                      'of \'PSF stars\' sets - stars used to derive PSF function')
    g_dao.add_argument('-c', '--all-stars-file', metavar='FILE', default=None,
                       help='all stars input file in one of daophot\'s formats (default: obtained by daophot FIND)')
    g_dao.add_argument('-D', '--daophot-opt', metavar='FILE', default=None,
                       help='daophot.opt file for daophot (default: internal)')
    g_dao.add_argument('--frames-av', metavar='n', type=int, default=1,
                       help='frames ave - parameter of daophot FIND when --all-stars-file not provided  (default: 1)')
    g_dao.add_argument('--frames-sum', metavar='n', type=int, default=1,
                       help='frames summed - parameter of daophot FIND when --all-stars-file not provided (default: 1)')
    g_dao.add_argument('-O', '--photo-opt', metavar='FILE', default=None,
                       help='photo.opt file for aperture photometry (default: none)')
    g_dao.add_argument('--photo-is', metavar='r', type=int, default=0,
                       help='PHOTOMETRY inner sky radius, overwrites photo.opt, (default: from --photo-opt or 35)')
    g_dao.add_argument('--photo-os', metavar='r', type=int, default=0,
                       help='PHOTOMETRY outher sky radius, overwrites photo.opt, (default: from --photo-opt or 50)')
    g_dao.add_argument('--photo-ap', metavar='r', type=int, default=[], nargs='+',
                       help='PHOTOMETRY apertures radius (up to 12), overwrites photo.opt, '
                            '(default: from --photo-opt or 8)')
    g_dao.add_argument('-l', '--psf-stars-file', metavar='FILE', default=None,
                       help='PSF candidates input file in one of daophot\'s formats, the result of algorithm is '
                            'a subset of those stars (default: obtained by daophot PICK)')
    g_dao.add_argument('-P', '--stars-to-pick', metavar='n', type=int, default=100,
                       help='number of stars to PICK as candidates when --stars-to-pick not provided (default: 100)')
    g_dao.add_argument('--faintest-to-pick', metavar='MAG', type=int, default=20,
                       help='faintest magnitude to PICK as candidates when --stars-to-pick not provided (default: 20)')
    g_dao.add_argument('-f', '--fine', action='store_true',
                       help='fine tuned PSF calculation (3 iter) for crowded fields, without this option no neighbours'
                            'subtraction will be performed (slower, usually not needed)')
    g_flt = parser.add_argument_group('Filtering training set stars',
                                      'Options controls which stars will be included in a training set -- whose '
                                      'residuals after the PSF fitting will be minimized by genetic algorithm')
    g_flt.add_argument('-x', '--include-psf', action='store_true',
                       help='NEW | include candidates of PSF stars in training set; by default PSF candidates are'
                            'excluded from calculation of minimized PSF error; may be included if there is low number'
                            'of stars')
    g_flt.add_argument('--max-ph-err', metavar='x', type=float, default=0.1,
                       help='threshold of aperture photometry error for training set stars; '
                            'stars for which aperture photometry (daophot PHOTO) error is greater than x '
                            'will be excluded and have no effect on minimization '
                            '(default 0.1)')
    g_flt.add_argument('--max-ph-mag', metavar='m', type=float, default=20,
                       help='threshold of aperture photometry magnitude of stars for for training set stars; '
                            '(analogous to --max-ph-err)stars for which aperture photometry (daophot PHOTO) magnitude '
                            'is greater than m (fainter than m) will be excluded form allstar run and have no effect '
                            'on quality measurement (default 20)')
    g_flt.add_argument('--clip-fit-sigmas', metavar='x', type=float, default=3.0,
                       help='NEW | threshold of errors from the preliminary PSF fitting (allstars) in sigmas above '
                            'mean error level for specific magnitude (a mean error level for the magnitude is fitted) '
                            'stars with preliminary PSF fitting error greater than mean_err(mag)+x*sigma will '
                            'be rejected from training set (default 3.0)')
    g_psf = parser.add_argument_group('Filtering candidates for PSF  stars',
                                      'Options controls which stars can be considered as PSF candidates. '
                                      'Filtering is applied to stars provided in --psf-stars-fil or found by '
                                      'daophot/PICK')
    g_psf.add_argument('--max-psf-err-mult', metavar='x', type=float, default=3.0,
                       help='threshold of PSF errors of PSF candidates - as multiplier of average PSF error; '
                            'candidates with PSF error (from daophot PSF command) greater than x*av_err will '
                            'be rejected (as well as other marked ? or * by daophot/PSF command)'
                            '(default 3.0)')
    g_psf.add_argument('--clip-psf-fit-sigmas', metavar='x', type=float, default=2.0,
                       help='NEW | threshold of errors from the preliminary PSF fitting (allstars) in sigmas above '
                            'mean error level for specific magnitude (a mean error level for the magnitude is fitted) '
                            'candidates with preliminary PSF fitting error greater than mean_err(mag)+x*sigma will '
                            'be rejected (default 2.0)')
    g_ga = parser.add_argument_group('Genetic algorithm parameters')
    g_ga.add_argument('-I', '--ga-init-prob', metavar='x', default=[0.3, 0.8], type=float, nargs='+',
                      help='what portion of candidates is used to initialize GA individuals;'
                           ' e.g. if there is 100 candidates, each of them will be '
                           ' chosen to initialize individual genome with probability x; '
                           ' in other words if x=0.3 first population in GA will contain'
                           ' individuals with around 30 stars each; '
                           ' two numbers (e.g. -I 0.3 0.8) indicates range: for each initial genome'
                           ' random number from that range will be taken (spreading number of initial stars)'
                           ' (default: 0.3 0.8)')
    g_ga.add_argument('-i', '--ga-max-iter', metavar='n', default=50, type=int,
                      help='maximum number of iterations of generic algorithm - generations (default: 50)')
    g_ga.add_argument('-n', '--ga-pop', metavar='n', default=80, type=int,
                      help='population size of GA (default: 80)')
    g_ga.add_argument('--ga-cross-prob', metavar='x', default=0.5, type=float,
                      help='crossover probability of GA (default: 0.5)')
    g_ga.add_argument('--ga-mut-prob', metavar='x', default=0.2, type=float,
                      help='mutation probability of GA - probability to became a mutant (default: 0.2)')
    g_ga.add_argument('--ga-mut-str', metavar='x', default=0.05, type=float,
                      help='mutation strength of GA - probability of every bit flip in mutant (default: 0.05)')
    g_cnt = parser.add_argument_group('gapick run control parameters')
    g_cnt.add_argument('-p', '--parallel', metavar='n', type=int, default=8,
                       help='how many parallel processes can be forked; '
                            'n=1 avoids parallelism (default: 8)')
    g_cnt.add_argument('-d', '--out-dir', metavar='output_dir', type=str, default='RESULTS',
                       help='output directory; directory will be created and result files will be stored there;'
                            ' directory should not exist or --overwrite flag should be set'
                            ' (default: RESULTS)')
    g_cnt.add_argument('-o', '--overwrite', action='store_true',
                       help='if directory specified by --out_dir parameter exists, '
                            'then ALL its content WILL BE DELETED')
    # g_cnt.add_argument('-C', '--checkpoint', metavar='file.chk', type=str, default=None,
    #                     help='restore evaluation from checkpoint; algorithm saves checkpoint.chk file every generation,'
    #                          ' which allows resuming evolution, even with another parameters')
    g_cnt.add_argument('-L', '--loglevel', metavar='level', default='info',
                       help='logging level: debug, info, warning, error, critical (default: info)')
    g_cnt.add_argument('-t', '--no-stdout', action='store_true',
                       help='suppress printing result (list of best choice of PSF stars) to stdout at finish')
    g_cnt.add_argument('-b', '--no-progress', action='store_true',
                       help='suppress showing progress bar')
    g_cnt.add_argument('-v', '--version', action='store_true',
                       help='show version and exit')
    g_cnt.add_argument('--demo', action='store_true',
                       help='use internal demo FITS file')
    __arg_parser_singlethon = parser
    return parser


def main(**kwargs):
    """Entry point for python script calls. Parameters identical to command line"""

    args = commons.bunch_kwargs(__arg_parser(), **kwargs)
    # call main routine - common form command line and python calls
    return __do(args)


def check_arguments(arg, parser=None):
    if len(arg.ga_init_prob) > 2:
        exit_error('--ga-init-prob parameter be one or two numbers', parser)
    if len(arg.ga_init_prob) == 1:
        arg.ga_init_prob = [arg.ga_init_prob[0], arg.ga_init_prob[0]]
    if max(arg.ga_init_prob) > 0.99 or min(arg.ga_init_prob) < 0.01:
        exit_error('--ga-init-prob must be in range [0.01 : 0.99]', parser)
    if arg.ga_init_prob[0] > arg.ga_init_prob[1]:
        exit_error('--ga-init-prob a b : a <= b must hold', parser)
    if not arg.demo and not arg.image:
        exit_error('FITS image parameter not provided (and no --version, --help or --demo)', parser)


def exit_error(message, parser):
    logging.error(message)
    if parser:
        parser.error()
        parser.exit(1)
    else:
        raise ValueError(message)


def info():
    """Prints commandline help message"""
    commons.info(__arg_parser())


def commandline_entry():
    # Entry point for command line
    __args = __arg_parser().parse_args()  # parse command line arguments
    if __args.version:
        print('astwro.tools ' + astwro.tools.__version__)
        exit()

    # Configure logging
    #    logging.basicConfig(format='[%(levelname)s] %(module)s: %(message)s', level=__args.loglevel.upper())
    logging.basicConfig(format='%(levelname)7s | %(message)s', level=__args.loglevel.upper())
    check_arguments(__args, __arg_parser())
    __stars = __do(__args)  # call main routine - common form command line and python calls
    if __stars is None:
        return 1
    if not __args.no_stdout:
        print('\n'.join(map(str, __stars.index)))
    return 0


if __name__ == '__main__':
    code = commandline_entry()
    exit(code)



# coding=utf-8
"""Differential photometry as in Honeycutt 1992PASP..104..435H


See Also
--------
dphot         : function calcualting differential photometry and light curves
dphot_filters : convenience function calling dphot for multiple filters
 """
from .dphot import *
from .phot_error import err_poly_fit, PhotError




#! /usr/bin/env python
# coding=utf-8
"""Copy this skeleton to start new tool
"""
from __future__ import print_function, division
# GLOBAL IMPORTS HERE
import __commons as commons


def __do(arg):
    """Main routine, common for command line, and python scripts call"""

    # refer to args: arg.foo
    # IMPLEMENT HERE
    return arg.bar + arg.foo


def __arg_parser():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=commons.version_string(),
        description='Sums foo and bar')
    parser.add_argument('foo', type=int,
                        help='some foo')
    parser.add_argument('--bar', type=int,
                        help='some bar')
    return parser


# Below: standard skeleton for astwro.tools
# CUSTOMIZE:
#   1. customize postional - obligatory arguments
#   2. decide whether to print results (__do return)

def main(positional1, positional2, **kwargs):
    """Entry point for python script calls. Parameters identical to command line"""
    # Extract default arguments from command line parser and apply kwargs parameters
    args = commons.bunch_kwargs(__arg_parser(), positional=[positional1, positional2], **kwargs)
    # call main routine - common form command line and python calls
    return __do(args)


def info():
    commons.info(__arg_parser())


def commandline_entry():
    # Entry point for command line
    __args = __arg_parser().parse_args()  # parse command line arguments
    if __args.version:
        print ('astwro.tools '+astwro.tools.__version__)
        exit()
    result = __do(__args)  # call main routine - common form command line and python calls
    print(result)
    return 0

if __name__ == '__main__':
    code = commandline_entry()
    exit(code)


# coding=utf-8
from __future__ import absolute_import, division, print_function
from astwro.exttools import Runner

class WCSUtilsRunner(Runner):
    def __init__(self, translator=None):
        # base implementation of __init__ calls `_reset` also
        self.translator = translator
        super(WCSUtilsRunner, self).__init__()
    pass


import unittest
from astwro.pydaophot import Daophot, Allstar

# TODO: rewrite tests for v 0.4+ and switch to pytest

class TestRunners(unittest.TestCase):

    @property
    def image(self):
        from astwro.sampledata import fits_image
        return fits_image()

    def test_sample_img_access(self):
        import os
        self.assertTrue(os.path.isfile(self.image))

    def test_execution_daophot(self):
        d = Daophot(image=self.image)
        d.run()
        x,y = d.ATtach_result.picture_size
        self.assertGreater(x, 0)
        self.assertGreater(y, 0)

    def test_image_substitution_and_find(self):
        d = Daophot()
        d.image = self.image
        r = d.FInd()
        print(d.output)
        self.assertGreater(r.stars, 1)

    def test_exception_in_PHotometry(self):
        d = Daophot(image=self.image, batch=True)
        d.FInd()
        with self.assertRaises(d.RunnerException):
            d.PHotometry(IS=0, OS=0, apertures=[8])

    #@unittest.skip('long run, un-skip to test allstar')
    def test_execution_allstar_pipeline(self):
        d = Daophot(batch=True)
        d.ATtach(self.image)
        d.FInd(1, 1)
        d.PHotometry(IS=35, OS=50, apertures=[8])
        d.PIck()
        d.PSf()
        d.run()
        a = Allstar(dir=d.dir, image=self.image)
        a.ALlstar(stars='i.nei')
        self.assertTrue(a.ALlstars_result.success)

    def test_execution_psf_pipeline(self):
        d = Daophot(image=self.image, batch=True)
        d.FInd(1, 1)
        d.PHotometry(IS=35, OS=50, apertures=[8])
        d.PIck()
        d.PSf()
        d.run()
        self.assertGreater(d.PSf_result.chi, 0)

    def test_sky_and_find(self):
        d = Daophot(image=self.image)
        s = d.SKy()
        f = d.FInd()
        self.assertAlmostEqual(f.sky, s.sky)
        self.assertAlmostEqual(f.skydev, s.skydev)
        self.assertAlmostEqual(f.mean, s.mean)
        self.assertAlmostEqual(f.median, s.median)
        self.assertAlmostEqual(f.pixels, s.pixels)

    def test_neda(self):
        d = Daophot(image=self.image)
        d.FInd(1, 1)
        d.PHotometry(IS=35, OS=50, apertures=[8])
        d.PIck()
        psf_stars = d.PIck_result.picked_starlist
        d.PSf()
        a = Allstar(dir=d.dir, image=self.image)
        a.ALlstar(stars='i.ap')
        self.assertTrue(a.ALlstars_result.success)
        d.NEda(IS=35, OS=50, apertures=[8,10,12.5],
               stars_id=psf_stars[:10])
        self.assertGreater(len(d.NEda_result.neda_starlist), 1)



suite = unittest.TestLoader().loadTestsFromTestCase(TestRunners)
if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)


# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import pytest
import os.path
from .. import daophot_bialkow
from astwro.utils import tmpdir
from astwro.sampledata import fits_image, coo_file, lst_file

#@pytest.mark.skip('Long test, excluded temporary')
def test_daophot_bialkow():
    d = tmpdir()
    ipath, iname = os.path.split(fits_image())
    daophot_bialkow.daophot_photometry(iname, all_stars=coo_file(), psf_stars=lst_file(),IMGPATH=ipath, OUTPATH=d.path)



"""
Module contains sample FITS file and other daophot files for testing
"""
def fits_image():
    """path of FITS image of NGC6871"""
    return __sampledata('NGC6871.fits')

def coo_file():
    """path of `coo` file for :py:func:`fits_image()`"""
    return __sampledata('NGC6871.coo')

def lst_file():
    """path of `lst` file for :py:func:`fits_image()` """
    return __sampledata('NGC6871.lst')

def ap_file():
    """path of `ap` file for :py:func:`fits_image()`"""
    return __sampledata('NGC6871.ap')

def psf_file():
    """path of `psf` file for :py:func:`fits_image()`"""
    return __sampledata('NGC6871.psf')

def als_file():
    """path of `als` file for :py:func:`fits_image()`"""
    return __sampledata('NGC6871.als')

def nei_file():
    """path of `nei` file for :py:func:`fits_image()`"""
    return __sampledata('NGC6871.nei')

def err_file():
    return __sampledata('i.err')

def head_file():
    """patch of sample ASCII fits header file"""
    return __sampledata('astrometry.head')

def sampledata_dir():
    import os
    return os.path.abspath(os.path.dirname(__file__))

def __sampledata(filename):
    import os
    return os.path.join(sampledata_dir(), filename)



#! /usr/bin/env python
# coding=utf-8

from __future__ import print_function, division

from scipy.stats import sigmaclip

from astwro.pydaophot import Daophot
from astwro.pydaophot import Allstar
from astwro.starlist import read_dao_file
from astwro.starlist import write_ds9_regions

# TODO: expand this script to (optionally) leave result files - make: it allstar runner

# image = args.image
# coo = args.coo
# lst = args.lst


def main(**kwargs):
    # 1 do daophot aperture and psf photometry and run allstar

    dp = Daophot(image=kwargs.image)
    dp.PHotometry(stars=kwargs.coo)
    dp.PSf(psf_stars=kwargs.lst)
    al = Allstar(dir=dp.dir)
    al.ALlstar()
    all_s = al.ALlstars_result
    print(sigmaclip(all_s.chi)[0].mean())
    all_s.hist('chi')

    # 2 write regions
    # if args.regions:
    #
    #     write_ds9_regions(coo_file+'.reg', )


if __name__ == '__main__':

    import argparse
    parser = argparse.ArgumentParser(
        description='Run daophot photometry and allstar. Find mean chi of allstar stars. '
                    '(chi is calculated by allstar for every star as '
                    '"the observed pixel-to-pixel scatter from the model image profile '
                    'DIVIDED BY the expected pixel-to-pixel scatter from the image profile"). '
                    'The same mean chi is used as function to be minimized by genetic '
                    'algorithm in genetic_psf_str.py. This script allows quick comparison '
                    'between different PSF stars sets.')
    parser.add_argument('image', type=str,
                        help='FITS image file')
    parser.add_argument('coo', type=str,
                        help='all stars list: coo file')
    parser.add_argument('lst', type=str,
                        help='PSF stars list: lst file')
    parser.add_argument('--reg', action='store_true',
                        help='create ds9 region files <name>.coo.reg and <name>.lst.reg')

    args = parser.parse_args()

    main(**args.__dict__)


# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import os
import random
from collections import namedtuple
from astwro.exttools import Runner
import astwro.starlist as sl

#TODO: Add convenient sorting for starlist to avoid manually sorting/writing: dp.write_starlist(sorted_stars, 'i.ap')

class DAORunner(Runner):
    """base for daophot package runners runner"""

    def __init__(self, dir=None, batch=False):
        super(DAORunner, self).__init__(dir=dir, batch=batch)

    def __deepcopy__(self, memo):
        return super(DAORunner, self).__deepcopy__(memo)

    # dao files management
    def apertures_file_push(self, src_path):
        """
        Copies aperture file photo.opt into working dir. File will be used by daophot
        :param str src_path: patch to src file
        :rtype: None
        """
        self.copy_to_runner_dir(src_path, 'photo.opt')

    def apertures_file_pull(self, dst_path = '.'):
        """
        Extracts current aperture file photo.opt from working dir.
        :param dst_path: destination
        :rtype: None
        """
        self.copy_from_runner_dir('photo.opt', dst_path)

    def apertures_file_create(self, apertures, IS, OS):
        """
        Creates photo.opt in daophot working dir from list
        :param list apertures: list of apertures A1,A2... e.g. [6.0,8.0,12.0]
        :param float IS: inner radius of sky annulus
        :param float OS: outer radius of sky annulus
        :rtype: None
        """
        assert len(apertures) > 0 and len(apertures) < 13
        self.rm_from_runner_dir('photo.opt')
        with open(os.path.join(self.dir, 'photo.opt'), 'w') as f:
            f.write(''.join('A{:1X}={:.2f}\n'.format(n+1, v) for n,v in zip(range(len(apertures)), apertures)))
            f.write('IS={:.2f}'.format(IS))
            f.write('OS={:.2f}'.format(OS))

    def read_starlist(self, filepath, **kwargs):
        # type: ([str], []) -> sl.StarList
        """
        Returns `StarList` object with stars extracted from daophot files
        :param [str] filepath: source file for starlist, if filename without path is provided, runner directory is assumed.
        :param  kwargs: additional parameters for extra processing in subclasses e.g. add_psf_errors=True
        :rtype: starlist.StarList
        """
        s = sl.read_dao_file(self.absolute_path(filepath))
        return self._process_starlist(s, **kwargs)

    def write_starlist(self, stars, filename=None, dao_file_type=None):
        # type: (sl.StarList, [str], namedtuple) -> str
        """
        Writes `StarList` object to file in runner directory 
        :param  sl.StarList stars: star list to be written
        :param  filename: name of file in runner directory, default: random name with extension '.stars'
        :return name of file in runner directory
        """
        if dao_file_type is None:
            dao_file_type = stars.DAO_type
        if filename is None:
            ext = dao_file_type.extension if dao_file_type else '.stars'
            filename = self._runner_dir_file_name(signature=random.random(), suffix=ext)
        sl.write_dao_file(stars, os.path.join(str(self.dir), filename), dao_type=dao_file_type)
        return filename


    def _prepare_input_file(self, data, default_dao_file_type=None):
        # check if input has a form of StarList
        if isinstance(data, sl.StarList):
            dao_type = default_dao_file_type if data.DAO_type is None else data.DAO_type
            data = self.write_starlist(data, dao_file_type=dao_type)
        return super(DAORunner, self)._prepare_input_file(data)

    def _process_starlist(self, s, **kwargs):
        return s




# coding=utf-8
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import logging
import numpy as np


def write_lc(hjd, lc, lc_e=None, ids=None, min_obs=1, prefix='lc_', suffix='.obs'):
    if ids is None:
        ids = range(lc.shape[0])
    masked = isinstance(lc, np.ma.MaskedArray)
    if masked:
        smask = lc.count(axis=1) >= min_obs
    else:
        smask = np.ones(lc.shape[0], dtype=bool)
    logging.info('Number of stars to save: {}'.format(np.count_nonzero(smask)))
    for i in smask.nonzero()[0]:
        fname = '{:s}{:05d}{:s}'.format(prefix, ids[i], suffix)
        logging.info('Processing star id: ',i)
        d = lc[i]
        if masked:
            m = ~d.mask
            t = hjd[m]
            e = lc_e[i][m] if lc_e is not None else None
            d = d.compressed()
        else:
            t = hjd
            e = lc_e[i] if lc_e is not None else None
        if lc_e is None:
            out = np.array([t, d]).T
        else:
            out = np.array([t, d, e]).T
        np.savetxt(fname, out, fmt='%15f')


def write_lc_filters(filters_masks, filternames, hjd, lc, lc_e=None, ids=None, min_obs=1, prefix='lc_', suffix='.obs'):
    if isinstance(filters_masks, dict):
        filters_masks = [filters_masks[f] for f in filternames]
    filters_masks = np.array(filters_masks, dtype=bool)

    for i, mask in enumerate(filters_masks):
        logging.info('Writing lc for Filter: {}'.format(filternames[i]))
        s = '_' + filternames[i] + suffix
        mhjd = hjd[mask]
        mlc = lc[:, mask]
        if lc_e is not None:
            mlc_e = lc_e[:, mask]
        else:
            mlc_e = None
        write_lc(mhjd, mlc, mlc_e, ids, min_obs, prefix, s)


# def write_lc(hjd, lc, lc_e=None, ids=None, min_obs=1, prefix='lc_', suffix='.obs'):
#     if ids is None:
#         ids = range(lc.shape[0])
#     smask = lc.count(axis=1) >= min_obs
#     logging.info('Number of stars to save: {}'.format(np.count_nonzero(smask)))
#     for i in smask.nonzero()[0]:
#         fname = '{:s}{:05d}{:s}'.format(prefix, ids[i], suffix)
#         logging.info('Processing star id: ',i)
#         with open(join(fname), 'w') as f:
#             for t, m in zip(hjd, lc[i]):
#                 if m is not np.ma.masked:
#                     print (t, m, file=f)



from __future__ import absolute_import, division, print_function

__metaclass__ = type
# see https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef
from os.path import join, basename, splitext
from astwro.exttools import Runner
from astwro.config import find_opt_file
from .Sextractor import SexResults


class ScampResults(object):

    def __init__(self, dir, results_filename):
        super(ScampResults, self).__init__()
        self.dir = dir
        self.header_file = results_filename
        self._header = None

    @property
    def fits_header(self):
        if self._header is None:
            import astropy.io.fits as pyfits
            self._header = pyfits.Header.fromtextfile(self.header_file, endcard=True)
        return self._header


class Scamp(Runner):
    """ Emmanuel BERTIN `scamp` runner

    Object of this class maintains single process of `scamp` and it's working directory.

    Note
    ----
    Very `alpha` stage, very limited


    Parameters
    ----------
    dir : str, TmpDir, optional
        Working directory for sextractor process
        Default - new TmpDir will be created and destroyed with Sextractor object
    conf : str, optional
        Patch to sextractor's configuration file, Default: package default file
    param : str, optional
        Patch to sextractor's parameters file (specified in conf), Default: package default file
    conv : str, optional
        Patch to sextractor's convolution file (specified in conf), Default: package default file
    """

    def __init__(self, dir=None, conf=None):
        # base implementation of __init__ calls `_reset` also
        super(Scamp, self).__init__(dir=dir, batch=False)
        self.conf_file = conf
        self.SCAMPresults = None
        self._update_executable('scamp')

    def _reset(self):
        super(Scamp, self)._reset()

    def __deepcopy__(self, memo):
        from copy import deepcopy
        new = super(Scamp, self).__deepcopy__(memo)

        return new

    def _pre_run(self, wait):
        super(Scamp, self)._pre_run(wait=wait)



    def _init_workdir_files(self, dir):
        super(Scamp, self)._init_workdir_files(dir)

    def __call__(self, src):
        """ Run scamp on crc catalog

        Parameters
        ----------
        src : str or SexResults
            Source catalog. Filename or SexResults

        Returns
        -------
        ScampResults object
        """
        if isinstance(src, SexResults):
            src = join(src.dir, src.stars_file)
        if self.conf_file is None:
            conf_file = find_opt_file('scamp.conf', package='sex')
        else:
            conf_file = self.conf_file

        out_file = splitext(basename(src))[0] # 'output1'


        self._prepare_input_file(conf_file, preservefilename = True)
        self._prepare_input_file(src, preservefilename=True)

        self.SCAMPresults = ScampResults(dir=self.dir.path,  results_filename=out_file)
        self.arguments = [basename(src), '-c', basename(conf_file)]
        self.run()
        return self.SCAMPresults






# coding=utf-8
from __future__ import absolute_import, division, print_function
import pytest
from astwro.phot.phot_error import PhotError


from astwro.phot import phot_error
import numpy as np

@pytest.fixture
def mag_err():
    import os
    import pickle
    pickelfile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'mag_err.npz')
    dic = np.load(pickelfile)
    return dic


def test_ph_error_tofit(mag_err):
    mag, err = mag_err['mag'], mag_err['mag_err']
    pe = PhotError(mag, err)
    assert pe.err_fitted.shape[0] > 0
    assert pe.err_means.shape[0] > 0
    assert pe.mag_means.shape[0] > 0
    assert len(pe.outlayers_mask) == len(mag)


def test_ph_error_custom_fn(mag_err):
    mag, err = mag_err['mag'], mag_err['mag_err']
    zzz = lambda x: np.percentile(x, 20) if len(x) > 0 else np.nan
    pe = PhotError(mag, err, meanfn=zzz, fitting_order=6, weighted_fit=True)
    assert pe.err_fitted.shape[0] > 0
    assert pe.err_means.shape[0] > 0
    assert pe.mag_means.shape[0] > 0
    assert len(pe.outlayers_mask) == len(mag)




# coding=utf-8
from __future__ import absolute_import, division, print_function


#from scipy import bin
import numpy as np
from scipy.stats import sigmaclip
from scipy.optimize import curve_fit
from cached_property import cached_property
import warnings



def err_poly_fit(mag, err):
    pe = PhotError(mag, err)
    return pe.fit


class PhotError(object):
    def __init__(self, mag, err, sigma_clip=3.0, bins='auto', fit_increasing_wing_only=True,
                 meanfn=np.mean, weighted_fit=True, fitting_order=6, fitplussigma=3.0):
        super(PhotError, self).__init__()
        self._mag = mag
        self._err = err
        self.sigma_clip = sigma_clip
        self.bins = bins
        self.fit_increasing_wing_only = fit_increasing_wing_only
        self.meanfn = meanfn
        self.weighted_fit = weighted_fit
        self.fitting_order = fitting_order
        self.fitplussigma = fitplussigma

    @cached_property
    def N(self):
        """Number of stars. Size of mag or err vector"""
        return len(self.mag)

    @cached_property
    def M(self):
        """Number of bins."""
        return len(self.mag_bin_counts)

    @cached_property
    def mag(self):
        return np.asanyarray(self._mag)

    @cached_property
    def err(self):
        return np.asanyarray(self._err)

    @cached_property
    def mag_idx_sorted(self):
        """Index sorting mag or err by mag. Size: N"""
        return np.argsort(self.mag)

    @cached_property
    def mag_sorted(self):
        """mag sorted. Size: N"""
        return self.mag[self.mag_idx_sorted]

    @cached_property
    def err_sorted(self):
        """err sorted by magnitude. Size: N"""
        return self.err[self.mag_idx_sorted]

    @cached_property
    def mag_histogram(self):
        """magnitude histogram: Pair: (values size M, edges size M+1)"""
        v, r = np.histogram(self.mag_sorted, bins=self.bins)
        return v, r

    @cached_property
    def mag_bin_counts(self):
        """Size: M"""
        return self.mag_histogram[0]

    @cached_property
    def mag_bin_edges(self):
        """Magnitudes of histogram bins edged. Size: M+1"""
        return self.mag_histogram[1]

    @cached_property
    def mag_bin_idx(self):
        """Indexes of magnitude histogram bins edged. Size: M+1"""
        divs = np.searchsorted(self.mag_sorted, self.mag_bin_edges[1:-1])
        return np.concatenate(([0], divs, [len(self.mag_sorted)]))

    @cached_property
    def mag_bin_idx_ranges(self):
        """Indexes ranges for bins of magnitude histogram. Size: M list of pairs"""
        #        return zip(self.mag_bin_idx, self.mag_bin_idx[1:])
        return list(zip(self.mag_bin_idx, self.mag_bin_idx[1:]))

    @cached_property
    def err_bin(self):
        return np.array([self.err_sorted[rlo:rhi] for rlo,rhi in self.mag_bin_idx_ranges])





    @cached_property
    def clipping(self):
        mask = np.ones_like(self.mag, dtype=bool)
        i = 0
        if self.sigma_clip:
            for rlo, rhi in self.mag_bin_idx_ranges:
                errors = self.err_sorted[rlo:rhi]
                if len(self.err_sorted[rlo:rhi]) > 2: # no sigmaclip for 0,1,2 element sets
                    _, elo, ehi = sigmaclip(self.err_sorted[rlo:rhi])
                    mask[rlo:rhi] = (elo <= self.err_sorted[rlo:rhi]) & (self.err_sorted[rlo:rhi] <= ehi)
#                    print(i, rlo, rhi, len(self.err_sorted[rlo:rhi]), mask[rlo:rhi].sum(), elo, ehi)
                i += 1
        return mask


    @cached_property
    def err_clipped(self):
        return np.array([self.err_sorted[rlo:rhi][self.clipping[rlo:rhi]] for rlo,rhi in self.mag_bin_idx_ranges])
    @cached_property

    def err_clipped_counts(self):
        return np.array([len(e) for e in self.err_clipped], dtype=int)

    @cached_property
    def mag_clipped(self):
#        for rlo,rhi in self.mag_bin_idx_ranges: #DEL
#            print(rlo, rhi, len(self.mag_sorted[rlo:rhi][self.clipping[rlo:rhi]]) )
        return np.array([self.mag_sorted[rlo:rhi][self.clipping[rlo:rhi]] for rlo,rhi in self.mag_bin_idx_ranges])

    @cached_property
    def err_means(self):
        with warnings.catch_warnings():  ## mean of empty slice allowed
            warnings.simplefilter("ignore", category=RuntimeWarning)
            ret = np.array([self.meanfn(e) for e in self.err_clipped])
        ret[self.err_clipped_counts == 0] = np.nan
        return ret

    @cached_property
    def err_means_stderr(self):
        with warnings.catch_warnings():  ## std of empty slice allowed
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return np.array([np.nanstd(e,ddof=1) for e in self.err_clipped])

    @cached_property
    def mag_means(self):
        with warnings.catch_warnings():  ## mean of empty slice allowed
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return np.array([np.nanmean(ms) for ms in self.mag_clipped])

    @cached_property
    def mean_mask(self):
        mask = np.isfinite(self.mag_means) \
               & np.isfinite(self.err_means_monotonized) \
               & np.isfinite(self.err_means_stderr)
        return mask

    @cached_property
    def err_means_monotonized(self):
        ret = self.err_means
        if self.fit_increasing_wing_only:
            ret = self.err_means.copy()
            leftcutout = np.nanargmin(self.err_means)
            ret [0:leftcutout] = self.err_means[leftcutout]
        return ret

    @cached_property
    def fit(self):
        return self._fit(self.err_means_monotonized[self.mean_mask])

    @cached_property
    def err_fitted(self):
        return self.evaluate_fit(self.mag_means)

    @cached_property
    def fit_plus_sigmas(self):
        return self._fit(self.err_means_monotonized[self.mean_mask]
                         + self.fitplussigma*np.nan_to_num(self.err_means_stderr[self.mean_mask]))

    @cached_property
    def err_fitted_plus_sigmas(self):
        return self.evaluate_fit_plus_sigmas(self.mag_means)

    @cached_property
    def outlayers_mask(self):
        limit = self.evaluate_fit_plus_sigmas(self.mag)
        return self.err > limit

    def evaluate_fit(self, x):
        return self._evaluate_fit(x, self.fit)

    def evaluate_fit_plus_sigmas(self, x):
        return self._evaluate_fit(x, self.fit_plus_sigmas)

    def _fit(self, y):
        x = self.mag_means[self.mean_mask]
        s = None
        if self.weighted_fit:
            s = (self.err_means_stderr[self.mean_mask])

        f = curve_fit(lambda t,a,b,c: a*np.exp(b*t)+c, x,y, sigma=s, absolute_sigma=True)
        return f

    def _evaluate_fit(self, x, fit):
        ret = np.zeros_like(x)
        return fit[0][0] * np.exp(fit[0][1] * x) + fit[0][2]

    # def _fit_poly(self, y):
    #     x = self.mag_means[self.mean_mask]
    #     w = None
    #     if self.weighted_fit:
    #         w = self.err_means_stderr[self.mean_mask] **-1.0  # for numpy polyfit ^-1 not ^-2 !!
    #         w[~np.isfinite(w)] = 0.0
    #     rank = min(len(x), 2)
    #     return np.polyfit(x,y, self.fitting_order, w=w)
    #
    # def _evaluate_fit_poly(self, x, fit):
    #     ret = np.zeros_like(x)
    #     for n, c in enumerate(fit[::-1]):
    #         ret += c * x**n
    #     return ret

    @cached_property
    def filtered_by_sigma_from_means(self):
        return self._filter(self.err_means)

    @cached_property
    def filtered_by_sigma_from_means(self):
        return self._filter(self.err_fitted)

    @cached_property
    def err_weighted_mean(self):
        """Weighted by flux mean of err, float"""
        w = 100 * (-self.mag / 5)
        return (self.err * w).sum() / w.sum()

    def _filter(self, err_bin_values):

        return


# coding=utf-8
from __future__ import absolute_import, division, print_function
import re
from astwro.exttools.output_processors import OutputBufferedProcessor
import astwro.starlist
from astwro.starlist import read_dao_file

__metaclass__ = type


# TODO: check for failure on all providers (raise_if_error)

# Daophot regexps:
#     for 'Command:' like
r_command = re.compile(r'Command:')
#     for 'Picture size:   1250  1150' like
r_pic_size = re.compile(r'(?<=Picture size:\s\s\s)([0-9]+)\s+([0-9]+)')
#     for options listing like FWHM OF OBJECT =     5.00   THRESHOLD (in sigmas) =     3.50
r_opt = re.compile(r'\b(\w\w)[^=\n]*=\s*(\-?[0-9]+\.[0-9]*)')
#     for SKy:
t_sky = (
    r'Sky mode and standard deviation = +(-?\d+\.\d*) +(-?\d+\.\d*)\n+ +'
    r'Clipped mean and median = +(-?\d+\.\d*)\s+(-?\d+\.\d*)\n +'
    r'Number of pixels used \(after clip\) = (\d*),?(\d+)')
r_sky = re.compile(t_sky)
#     for FInd:
t_find = t_sky + (
    r'\n +Relative error = +(-?\d+\.\d*)(?:\n.*)+\s'
    r'(\d+) stars'
)
r_find = re.compile(t_find)
# r_find = re.compile(
#     r'Sky mode and standard deviation = +(-?\d+\.\d*) +(-?\d+\.\d*)\n+ +'
#     r'Clipped mean and median = +(-?\d+\.\d*)\s+(-?\d+\.\d*)\n +'
#     r'Number of pixels used \(after clip\) = (\d*),?(\d+)\n +'
#     r'Relative error = +(-?\d+\.\d*)(?:\n.*)+\s'
#     r'(\d+) stars'
# )
#    for PHotometry
r_phot = re.compile(r'Estimated magnitude limit \(Aperture 1\): +(-?\d+\.\d*) +\+- +(-?\d+\.\d*) +per star')
#    for PIck
r_pick = re.compile(r'(\d+) +suitable candidates')
#    for PSf
r_psf = re.compile(r'Chi {4}Parameters...\n>* +(-?\d+\.\d*) +(-?\d+\.\d*) +(-?\d+\.\d*)')
r_psf_errors = re.compile(r' (\d+) +(\d+.\d+) ([ ?*])')
r_psf_failed_to_converge = re.compile(r'Failed to converge')
#    for GRoup
r_grp = re.compile(r'(\d+) +(\d+)')
r_grp_summ = re.compile(r'(\d+) +stars in (\d+) +groups.')

# Allstar regexp
r_alls_separator = re.compile(r'Input image name:')
r_alls_opt = r_opt
r_alls = re.compile(r'(\d+) +(\d+) +(\d+) +(\d+).*\n')


#  DAOPHOT

class DaophotCommandOutputProcessor(OutputBufferedProcessor):

    def _is_last_one(self, line, counter):
        return r_command.search(line) is not None

class DPOP_ATtach(DaophotCommandOutputProcessor):
    """Results of `ATTACH` daophot command"""
    @property
    def picture_size(self):
        """tuple with (x,y) size of pic returned by 'ATTACH' """
        buf = self.get_buffer()
        match = r_pic_size.search(buf)
        if match is None:
            raise Exception('daophot failed to attach image file. Output buffer:\n ' + buf)
        return int(match.group(1)), int(match.group(2))

    def raise_if_error(self):
        _ = self.picture_size


class DPOP_OPtion(DaophotCommandOutputProcessor):
    """Results of `OPTION` daophot command, or initial daophot options"""
    __options = None

    @property
    def options(self):
        """Dictionary of options: XX: 'nnn.dd'
           keys are two letter option names
           values are strings"""
        if self.__options is None:
            buf = self.get_buffer()
            match = dict(r_opt.findall(buf))
            if 'RE' not in match:  # RE not found - sth wrong, found, suppose is OK
                raise Exception('daophot failed to present options. Output buffer:\n ' + buf)
            self.__options = match
        return self.__options

    def get_option(self, key):
        """single option"""
        return float(self.options[key[:2].upper()])

    def raise_if_error(self):
        _ = self.options


class DpOp_SKy(DaophotCommandOutputProcessor):
    """Results of `SKY` daophot command"""

    def __init__(self, prev_in_chain=None):
        self.regexp = r_sky
        self.__data = None
        super(DpOp_SKy, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def data(self):
        if self.__data is None:
            buf = self.get_buffer()
            match = self.regexp.search(buf)
            if match is None:
                raise Exception('daophot SKY or FIND output doesnt match regexp for that command:'
                                ' error (or regexp is wrong). Output buffer:\n ' + buf)
            self.__data = match.groups()
        return self.__data

    @property
    def sky(self):
        """Sky estimation"""
        return float(self.data[0])

    @property
    def skydev(self):
        """Standart deviation of :var sky"""
        return float(self.data[1])

    @property
    def mean(self):
        """Mean of image"""
        return float(self.data[2])

    @property
    def median(self):
        """Median of image"""
        return float(self.data[3])

    @property
    def pixels(self):
        """Number of analyzed pixels"""
        t = self.data[4]
        if t is None or t == '':
            t = 0
        else:
            t = int(t) * 1000
        return int(self.data[5]) + t



class DpOp_FInd(DpOp_SKy):
    """Results of `FIND` daophot command (extends SKY)"""
    def __init__(self, prev_in_chain=None, starlist_file=None):
        self.__starlist = None
        self.starlist_file = starlist_file  #: Patch to output file with found stars
        super(DpOp_FInd, self).__init__(prev_in_chain=prev_in_chain)
        self.regexp = r_find

    @property
    def found_starlist(self):
        """StarList with found stars"""
        if self.__starlist is None and self.starlist_file:
            self.__starlist = read_dao_file(self.starlist_file)
        return self.__starlist

    @property
    def err(self):
        """Error estimation"""
        return float(self.data[6])

    @property
    def stars(self):
        """Number of found stars"""
        return int(self.data[7])

class DpOp_PHotometry(DaophotCommandOutputProcessor):
    """Results of `PHOTOMETRY` daophot command"""
    def __init__(self, prev_in_chain=None, photometry_file=None):
        self.__data = None
        self.__starlist = None
        self.photometry_file = photometry_file  #: Patch to output file with aperture photometry
        super(DpOp_PHotometry, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def data(self):
        if self.__data is None:
            buf = self.get_buffer()
            match = r_phot.search(buf)
            if match is None:
                raise Exception('daophot PH output doesnt match regexp r_phot:'
                                ' error (or regexp is wrong). Output buffer:\n ' + buf)
            self.__data = match.groups()
        return self.__data

    @property
    def mag_limit(self):
        return float(self.data[0])

    @property
    def mag_err(self):
        return float(self.data[1])

    @property
    def photometry_starlist(self):
        """StarList with photometry

        :rtype: astwro.starlist.StarList 
        """
        if self.__starlist is None and self.photometry_file:
            self.__starlist = read_dao_file(self.photometry_file)
        return self.__starlist


class DpOp_PIck(DaophotCommandOutputProcessor):
    """Results of `PICK` daophot command"""
    def __init__(self, prev_in_chain=None, picked_stars_file=None):
        self.__stars = None
        self.__starlist = None
        self.picked_stars_file = picked_stars_file  #: Patch to output file with picked stars
        super(DpOp_PIck, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def stars(self):
        """Number of picked stars"""
        if self.__stars is None:
            buf = self.get_buffer()
            match = r_pick.search(buf)
            if match is None:
                raise Exception('daophot PIck output doesnt match regexp r_pick:'
                                ' error (or regexp is wrong). Output buffer:\n ' + buf)
            self.__stars = int(match.group(1))
        return self.__stars

    @property
    def picked_starlist(self):
        """StarList with picked stars"""
        if self.__starlist is None and self.picked_stars_file:
            self.__starlist = read_dao_file(self.picked_stars_file)
        return self.__starlist



class DpOp_PSf(DaophotCommandOutputProcessor):
    """Results of `PSF` daophot command"""
    def __init__(self, prev_in_chain=None, psf_file=None, nei_file=None, err_file=None):
        self.__data = None
        self.__errors = None
        self.__neilist = None
        self.psf_file = psf_file  #: Patch to output file with PSF function
        self.nei_file = nei_file  #: Patch to output neighbours file 
        self.err_file = err_file  #: Patch to output errors file
        super(DpOp_PSf, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def nei_starlist(self):
        """StarList with neighbours stars"""
        if self.__neilist is None and self.nei_file:
            self.__neilist = read_dao_file(self.nei_file)
        return self.__neilist

    @property
    def converged(self):
        """False if daophot PSF routine does not produced result, e.g. 'Failed to converge'"""
        return self.__get_data() is not None

    @property
    def errors(self):
        """StarList (pandas.DataFrame) of PSF stars with errors and flags ('?', '*' or ' ')
        
        This information is identical to i.err file, but obtained directly from daophot output"""
        if self.__errors is None:
            buf = self.get_buffer()
            lst = [(int(star), float(err), flag) for star, err, flag in r_psf_errors.findall(buf)]
            sl = astwro.starlist.StarList(lst)
            sl.columns = ['id','psf_err','flag']
            sl.index = sl.id
            self.__errors = sl
        return self.__errors

    @property
    def data(self):
        d = self.__get_data()
        if d is None:
            raise Exception('daophot PSF output doesnt match regexp r_psf:'
                            ' PDF failed to converge?. Output buffer:\n ' + self.get_buffer())
        return d

    @property
    def chi(self):
        """Chi error estimation"""
        return float(self.data[0])

    @property
    def hwhm_xy(self):
        """tuple (x,y) of halfwidth's of PSF function"""
        return float(self.data[1]), float(self.data[2])

    def __get_data(self):
        if self.__data is None:
            buf = self.get_buffer()
            match = r_psf.search(buf)
            if match is not None:
                self.__data = match.groups()
        return self.__data

    def raise_if_error(self):
        if not self.converged:
            raise Exception('PSF not converged')


class DpOp_SUbstar(DaophotCommandOutputProcessor):
    """Results of `SUBSTAR` daophot command"""
    def __init__(self, prev_in_chain=None, subtracted_image_file=None):
        self.subtracted_image_file = subtracted_image_file  #: Patch to output fits image
        super(DpOp_SUbstar, self).__init__(prev_in_chain=prev_in_chain)

class DpOp_GRoup(DaophotCommandOutputProcessor):
    """Results of `GROUP` daophot command"""
    def __init__(self, prev_in_chain=None, groups_file=None):
        self.__data = None
        self.__grouphist = None
        self.groups_file = groups_file  #: Patch to output file with groups
        super(DpOp_GRoup, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def groups_histogram(self):
        """List of tuples: (size_of_group, number_of_groups)"""
        if self.__grouphist is None:
            buf = self.get_buffer()
            self.__grouphist = [(int(size), int(count)) for size, count in r_grp.findall(buf)]
        return self.__grouphist

    @property
    def data(self):
        if self.__data is None:
            buf = self.get_buffer()
            match = r_grp_summ.search(buf)
            if match is not None:
                self.__data = match.groups()
        return self.__data

    @property
    def stars(self):
        """Number of grouped stars reported by daophot `GROUP` command"""
        return float(self.data[0])

    @property
    def groups(self):
        """Number of groups"""
        return float(self.data[1])


class DpOp_NEda(DaophotCommandOutputProcessor):
    def __init__(self, prev_in_chain=None, neda_file=None):
        self.neda_file = neda_file #: Patch to output file with NEDA photometry
        self.__nedalist = None
        super(DpOp_NEda, self).__init__(prev_in_chain=prev_in_chain)

    @property
    def neda_starlist(self):
        """stars list with neda photometry"""
        if self.__nedalist is None and self.neda_file:
            self.__nedalist = read_dao_file(self.neda_file)
        return self.__nedalist



class DpOp_SOrt(DaophotCommandOutputProcessor):
    pass


#  ALLSTARS

class AsOp_opt(OutputBufferedProcessor):
    __options = None

    def _is_last_one(self, line, counter):
        return r_alls_separator.search(line) is not None

    @property
    def options(self):
        """returns dictionary of options: XX: 'nnn.dd'
           keys are two letter option names
           values are strings"""
        if self.__options is None:
            buf = self.get_buffer()
            match = dict(r_opt.findall(buf))
            if 'WA' not in match:  # WA not found - sth wrong, found, suppose is OK
                raise Exception('allstar failed to present options. Output buffer:\n ' + buf)
            self.__options = match
        return self.__options

    def get_option(self, key):
        return float(self.options[key[:2].upper()])

    def raise_if_error(self):
        _ = self.options

class AsOp_result(OutputBufferedProcessor):
    def __init__(self, prev_in_chain=None, profile_photometry_file=None, subtracted_image_file=None):
        self.__stars = None
        self.__als_stars = None
        self.profile_photometry_file = profile_photometry_file
        self.subtracted_image_file = subtracted_image_file
        super(AsOp_result, self).__init__(prev_in_chain=prev_in_chain)

    def _is_last_one(self, line, counter):
        return False

    @property
    def als_stars(self):
        # type: () -> {astwro.starlist.StarList}
        """StarList of stars with profile photometry results        """
        if self.__als_stars is None and self.profile_photometry_file:
            self.__als_stars = read_dao_file(self.profile_photometry_file)
        return self.__als_stars


    @property
    def stars_no(self):
        # type: () -> (int,int)
        """tuple: (converged_stars, disappeared_stars)"""
        if self.__stars is None:
            buf = self.get_buffer()
            match = r_alls.findall(buf)
            if len(match) < 1:
                self.__stars = 0, 0  # error or sth
            else:
                res = match[-1][-2:]  # last occurrence, two last values
                self.__stars = int(res[1]), int(res[0])
        return self.__stars

    def raise_if_error(self):
        if not self.stars_no[0] > 0:
            raise Exception('allstar: No stars')



from __future__ import absolute_import, division, print_function

import logging

# Override to your logger for logging, e.g:
#      pydaophot.logger = logging.getLogger('root').getChild('pydaophot')
logger = logging.getLogger('astwro')
# logger.addHandler(logging.NullHandler())



# coding=utf-8
from __future__ import absolute_import, division, print_function
from astropy.coordinates import SkyCoord
import numpy as np
from os import path
from io import StringIO
from .WCSUtilsRunner import WCSUtilsRunner



class XY2Sky(WCSUtilsRunner):
    def __init__(self, translator=None):
        # base implementation of __init__ calls `_reset` also
        super(XY2Sky, self).__init__(translator=translator)
        self._update_executable('xy2sky')

    def __call__(self, x, y, translator=None):
        """
        Parameters
        ----------
        x,y : float or array_like

        Returns
        -------
        SkyCoord object
        """
        if translator is None:
            translator = self.translator

        assert translator is not None, 'Set fits file with WCS as "translator"'

        translator = path.abspath(translator)
        self._prepare_input_file(translator, preservefilename = True)
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        xyfile = path.join(self.dir.path, 'xy')
        np.savetxt(xyfile, np.column_stack([x,y]), fmt='%12.6f')


        self.arguments = ['-dn', '6', translator, '@xy']
        self.run()
        ra,dec = np.loadtxt(StringIO(self.output), usecols=[0,1], unpack=True)
        return  SkyCoord(ra,dec, unit='deg')



# coding=utf-8
from __future__ import absolute_import, division, print_function

__metaclass__ = type


class Simulated_ri_rHa_Colors(object):
    """
    Simulated patches on (r' - i') -> (r' - Ha') diagram

    For different reddening and spectral types. Implemented after
    G. Barentsen et al. *T Tauri candidates in IC 1396 using IPHAS*
    [2011  MNRAS 415, 103–132]

    Parameters
    ----------
    reddening : float
        E(B-V)
    """
    _table = None

    def __init__(self, reddening=0.0):
        super(Simulated_ri_rHa_Colors, self).__init__()
        self._reddening = reddening
        self._r_i = None
        self._r_Ha = None

    @property
    def reddening(self):
        return self._reddening

    @reddening.setter
    def reddening(self, value):
        if value != self._reddening:
            self._r_i = None
            self._r_Ha = None
            self._reddening = value

    @property
    def r_i(self):
        if self._r_i is None:
            t = self._get_table()
            n = int(self._reddening)
            self.r_i = t.r_i[:, :, n] + self.reddening * (t.r_i[:, :, n+1] - t.r_i[:, :, n])
        return self.r_i

    @property
    def r_Ha(self):
        if self._r_Ha is None:
            t = self._get_table()
            n = int(self._reddening)
            self._r_Ha = t.r_Ha[:, :, n] + self.reddening * (t.r_Ha[:, :, n+1] - t.r_Ha[:, :, n])
        return self._r_Ha




    @classmethod
    def _get_table(cls):
        if cls._table is None:
            import os
            import pickle
            pickelfile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'iphas_simulated_paths.pkl')
            with open(pickelfile, 'rb') as f:
                cls._table = pickle.load(f)
        return cls._table
